# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Tuple
from .._ops import *
from datetime import datetime, timezone, timedelta
from functional import seq
from ..internals import is_delegate, root_node_blob_index, BlobType
from .._core import *
from .. import internals
from ..VT import TX,String

def yo_implementation(x, display=True):
    import inspect
    from ..fx.fx_types import FXElement, _group_types
    from ..fx import _effect_handlers
    from ..graph_slice import GraphSlice
    
    if display:
        import sys
        file = sys.stdout
    else:
        import io
        file = io.StringIO()

    if (type(x) == EZefRef or type(x) == ZefRef) and BT(x) == BT.TX_EVENT_NODE:
        print(tx_view(x), file=file)
    elif type(x) == EZefRef:
        if is_delegate(x):
            print("\n\n\n**************  delegate EZefRef ********************\n\n", file=file)
        else:
            print(eternalist_view(x), file=file)
    elif type(x) == ZefRef:
        if is_delegate(x):
            print("\n\n\n**************  delegate ZefRef ********************\n\n", file=file)
        else:
            print(eternalist_view(x), file=file)
    elif type(x) == Graph:
        print(graph_info(x), file=file)
    elif "pyzef.Graph" in str(type(x)):
        # This is because of monkeypatching
        print(graph_info(x), file=file)

    elif isinstance(x, GraphSlice):
        return yo_implementation(x | to_tx | collect, display)

    elif type(x) == ZefOp:
        if len(x.el_ops) == 1:
            from .dispatch_dictionary import _op_to_functions
            if len(x.el_ops) == 1:
                f = _op_to_functions[x.el_ops[0][0]][0]
                doc = inspect.getdoc(f)
                doc = doc if doc else f"No docstring found for given {x}!"
                print(doc, file=file)
        else:
            from .yo_ascii import make_op_chain_ascii_output
            print(make_op_chain_ascii_output(x), file=file)
    
    elif type(x) == FXElement:
        handler = _effect_handlers.get(x.d, None)
        if handler:
            doc = inspect.getdoc(handler)
            doc = doc if doc else f"No docstring found for given {x}!"
            print(doc, file=file)
        else:
            print(x, file=file)
    elif type(x) in _group_types:
        summary = [f"Summary of {x._name} group Class:"]
        for k,v in _effect_handlers.items():
            if k[0] == x._name and k[1] in x.__dir__()| filter[lambda x: not x.startswith("_")] | collect:
                summary.append(f"{fill_str_to_length(k[1], 20)} =>  {v}")
        print('\n'.join([str(i) for i in summary]), file=file)
        
    else:
        print(x, file=file)

    if display:
        return None
    else:
        return file.getvalue()

def yo_type_info(op, curr_type):
    return String



##############################################################
# * Implementation details below
#------------------------------------------------------------
def tx_view(zr_or_uzr) -> str:

    print(zr_or_uzr)
    print('!!!!!!!!!!!!!!!!!!!!!!!!!!!!!')
    print(zr_or_uzr)
    print(type(zr_or_uzr))
    print('!!!!!!!!!!!!!!!!!!!!!!!!!!!!!')
    # return 555

    uzr = to_ezefref(zr_or_uzr)

    def value_assigned_string_view(lst):
        def get_aet_values(x):
            return f"    {fill_str_to_length(value_previous_of_aet(x, uzr), 25)} ({uid(x)})        {value_of_aet_at_tx(x, uzr)}"
        return f"{len(lst)}x:     {zr_type(lst[0])}\n" + "\n".join(seq(lst).map(get_aet_values)) + "\n\n"

    def instantiated_or_terminated_string_view(lst):
        return (f"{len(lst)}x:     ({zr_type(lst[0])})\n"
                + "\n".join(seq(lst).map(lambda x: f"    ({uid(x)})"))
                + "\n\n")

    def tx_block_view(uzrs, fn):
        return ("".join(
                seq(uzrs)
                .group_by(lambda z: zr_type(z))
                .map(lambda x: x[1])
                .map(fn)))

    return f"""
======================================================================================================================
================================================== Transaction View ==================================================
================================= Transaction Time: {readable_datetime_from_tx_uzr(zr_or_uzr)} ==================================
======================================================================================================================

uid:                    {uid(uzr)}
blob index:             {index(uzr)}
current owning graphs:  {uid(Graph(uzr))} {f", name tags: ({','.join(Graph(uzr).graph_data.tag_list)})"
    if Graph(uzr).graph_data.tag_list else ""}
total affected:         {length(uzr | affected)}
total instantiations:   {length(uzr | instantiated)}
total assignments:      {length(uzr | value_assigned)}
total terminations:     {length(uzr | terminated)}
\n^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ Instantiations ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
{tx_block_view(uzr | instantiated | collect, instantiated_or_terminated_string_view)} 
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ Value Assignments ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
{tx_block_view(uzr | value_assigned | collect, value_assigned_string_view)} 
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ Terminations ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
{tx_block_view(uzr | terminated | collect, instantiated_or_terminated_string_view)} 
"""


def indent_lines(s: str, indent_amount=4):
    m = ' ' * indent_amount
    return m + s.replace('\n', '\n' + m)


def fill_str_to_length(s: str, target_length):
    if s is None:
        s = '/'
    return s + ' ' * max(0, target_length - len(s))

def maybe_elide(s: str, max_lines=10):
    # Shorten to maximum number of lines if too big and short_output mode is on
    if not zwitch.short_output():
        return s

    if s.count('\n') <= max_lines:
        return s

    def indexnth(s, c, n, ind=0):
        ind = s.index(c, ind)
        if n == 1:
            return ind
        else:
            return indexnth(s, c, n-1, ind+1)
    def rindexnth(s, c, n, ind=None):
        ind = s.rindex(c, 0, ind)
        if n == 1:
            return ind
        else:
            # Note: end=ind means it doesn't include ind in the next search
            return rindexnth(s, c, n-1, ind)

    start = indexnth(s, '\n', max_lines // 2)
    end = rindexnth(s, '\n', max_lines // 2)

    return s[:start] + "\n......<snip>......\n" + s[end+1:]

    


def zr_type(uzr):
    return repr(rae_type(uzr))


def value_previous_of_aet(aet, tx) -> str:
    try:
        if BT(aet) == BT.ATOMIC_ENTITY_NODE:
            prev_tx = tx << BT.NEXT_TX_EDGE | collect
            val = aet | value[prev_tx] | collect  # If prev_tx == BT.ROOT_NODE it will be caught by except clause
            return f" [previous val: {'/' if not val else str(val)}]"
        else:
            return ''
    except:
        return str(f" [previous val: NA]")


def value_of_aet_at_tx(aet, tx) -> str:
    latest_or_current = "latest" if (tx is now or len(tx | Outs[BT.NEXT_TX_EDGE] | collect) == 0) else "current"
    try:
        if tx is now:
            tx = Graph(aet) | now | collect
        if BT(aet) == BT.ATOMIC_ENTITY_NODE:
            val = aet | value[tx] | collect 
            return f" [{latest_or_current} val: {'/' if val is None else str(val)}]"
        else:
            return ''
    except:
        return str(f" [{latest_or_current} val: NA]")


def readable_datetime_from_tx_uzr(uzr_tx) -> str:
    uzr_tx = to_ezefref(uzr_tx)
    if index(uzr_tx) == root_node_blob_index():
        return '/'
    return str(time_slice(uzr_tx)) + ': ' + str(uzr_tx | time | collect)

    # time_stamp_here = unix_time(uzr_tx)
    # return datetime.fromtimestamp(time_stamp_here)\
    #     .astimezone(timezone(timedelta(hours=8)))\
    #     .strftime(f"TimeSlice {fill_str_to_length(str(time_slice(uzr_tx).value), 6)} %d.%m.%Y %H:%M:%S (SG), {time_stamp_here}") \
    #     if index(uzr_tx) != internals.root_node_blob_index() else "/"


def eternalist_view(zr_or_uzr) -> str:
    reference_frame = zr_or_uzr | frame | collect if isinstance(zr_or_uzr, ZefRef) else Graph(zr_or_uzr) | now | collect
    reference_tx = reference_frame | to_tx | collect
    uzr = to_ezefref(zr_or_uzr)
    return f"""
======================================================================================================================
================================================== {'Historical ' if isinstance(zr_or_uzr, ZefRef) else 'Most Recent'} View ==================================================
===================================== {'Seen from' if isinstance(zr_or_uzr, ZefRef) else 'Up to timeslice'}: {readable_datetime_from_tx_uzr(reference_tx)}  =================================
======================================================================================================================

uid:                    {uid(uzr)}
blob index:             {index(uzr)}
type:                   {zr_type(uzr)}
current owning graphs:  {Graph(uzr) | uid | collect} {f", name tags: ({','.join(Graph(uzr).graph_data.tag_list)})"
    if Graph(uzr).graph_data.tag_list else ""}
other graphs viewing:   /
instantiation:          {readable_datetime_from_tx_uzr(uzr | instantiation_tx | collect)}
termination:            {readable_datetime_from_tx_uzr(uzr | termination_tx | collect) if not exists_at(uzr, reference_frame) else '/'}
{src_dst_info(uzr) if BT(uzr) == BT.RELATION_EDGE else ''}
""" + \
           indent_lines(relations_view(zr_or_uzr)) + "\n" + \
           indent_lines(maybe_elide(timeline_view(zr_or_uzr), 30))


def src_dst_info(uzr) -> str:
    return f"""
    ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ Source and Target ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    ({uzr | source | uid | collect}: {zr_type(uzr | source | collect)})-----(z)----->({uzr | target | uid | collect}: {zr_type(uzr | target | collect)})
"""


def relations_view(zr_or_uzr) -> str:
    reference_frame = zr_or_uzr | frame | collect if isinstance(zr_or_uzr, ZefRef) else Graph(zr_or_uzr) | now | collect
    reference_tx = reference_frame | to_tx | collect
    def compose_triple_name(rel, z):
        if to_ezefref(rel | source) == to_ezefref(z):
            return ('outs', zr_type(rel), zr_type(rel | target | collect))
        else:
            return ('ins', zr_type(rel), zr_type(rel | source | collect))

    edges_to_show = ( 
        (zr_or_uzr | in_rels[RT], zr_or_uzr | out_rels[RT]) 
        | concat 
        | (identity if isinstance(zr_or_uzr, ZefRef) else filter[is_zefref_promotable])
        | collect
        )                     

    list_of_and_and_outs = (seq(edges_to_show)
                            .group_by(lambda rel: compose_triple_name(rel, zr_or_uzr))
                            .sorted(lambda rel_group: len(rel_group[1])))  # the second element is the list of zefrefs

    def wrap_to_fixed_length(in_str, total_length=51):
        m = total_length - len(in_str)
        return f"{'-' * round(m / 2 - 0.1)}-({in_str})-{'-' * round(m / 2 + 0.1)}"  # hacky wey to get ceil / floor

    def single_relation_type_view(triple_type_and_list) -> str:
        l_symb = '<' if triple_type_and_list[0][0] == 'ins' else '-'
        r_symb = '-' if triple_type_and_list[0][0] == 'ins' else '>'
        newline = '\n'
        return f"""\
{len(triple_type_and_list[1])}x:     (z:{zr_type(zr_or_uzr)}) {l_symb}---{wrap_to_fixed_length(triple_type_and_list[0][1])}--{r_symb} ({triple_type_and_list[0][2]})
{maybe_elide(''.join([f"            (z) {l_symb}---({ro | to_ezefref | uid | collect})---{r_symb} ({ro | to_ezefref | (source if l_symb == '<' else target) | uid | collect}{value_of_aet_at_tx(ro | (source if l_symb == '<' else target) | collect,  reference_tx | to_ezefref | collect)}){newline}" for ro in triple_type_and_list[1]]))}
"""

    return f"""\
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ Incoming & Outgoing Relations ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

{''.join([single_relation_type_view(trip) for trip in list_of_and_and_outs])}
"""


def timeline_view(zr_or_uzr) -> str:
    uzr = to_ezefref(zr_or_uzr)
    reference_tx = to_tx(frame(zr_or_uzr)) if isinstance(zr_or_uzr, ZefRef) else Graph(zr_or_uzr) | now | to_tx | collect

    # Function: visually list all transactions of affecting a certain entity
    # traverse the low level graph here to gather instantiation, termination and value assignment edges
    def generate_low_level_description(low_level_edge_uzr):
        if BT(low_level_edge_uzr) == BT.INSTANTIATION_EDGE:
            return f'=========  <---- Instantiation'
        # if low_level_edge_uzr | BT == BT.CLONE_REL_ENT_EDGE:
        #     return f'<<<<<<<<<  <---- Cloned: ....'
        if BT(low_level_edge_uzr) == BT.ATOMIC_VALUE_ASSIGNMENT_EDGE:
            return f'    x      <---- Value assignment: {value_of_aet_at_tx(uzr,low_level_edge_uzr | source | collect)}'  # get the value: pass the respective tx
        if BT(low_level_edge_uzr) == BT.TERMINATION_EDGE:
            return f'=========  <---- Termination'
        raise Exception("To be fixed with new lineage system")

    # Function: visually list all details about a single relation.
    # Output is affected by the "in" or "out" direction and also the state of the rt_uzr.
    def generate_rt_description(rt_uzr: EZefRef, direction: str, state: str) -> str:
        connected_et_or_aet = rt_uzr | source | collect if direction == "in" else rt_uzr | target | collect 
        sign = "<<-" if direction == "in" else "->>"
        instantiation_or_termination_tx = rt_uzr | instantiation_tx | collect if state == "Instantiated" else rt_uzr | termination_tx | collect
        return (f'    |            {fill_str_to_length(f"{state} Relation: ",50)}[{readable_datetime_from_tx_uzr(instantiation_or_termination_tx)}]\n'
                f'    |            {fill_str_to_length(f"({zr_type(rt_uzr)})",50)}({zr_type(connected_et_or_aet)}) {value_of_aet_at_tx(connected_et_or_aet,reference_tx|to_ezefref | collect)}\n'
                f'    |{sign if direction == "in" else "-"}---------({rt_uzr|uid|collect})---------------{sign if direction == "out" else "-"}({connected_et_or_aet | uid | collect})')

    # Appends directed relations to the all_edges list which is in the scope of this function
    def add_directed_rt_to_list(edges: EZefRefs, direction: str) -> None:
        for e in edges:
            all_edges.append((e, direction, "Instantiated"))
            if e | termination_tx | BT | collect == BT.TX_EVENT_NODE:
                all_edges.append((e, direction, "Terminated"))

    # Returns the time of the a transaction depending on the BT type and the edge_state.
    def edge_time_by_type(edge_info_triple: tuple) -> Time:
        edge, _, edge_state = edge_info_triple
        if BT(edge) == BT.RELATION_EDGE:
            return (edge | instantiation_tx | time | collect) if edge_state == "Instantiated" else (edge | termination_tx | time | collect)
        return edge | source | time | collect

    reference_tx_timeslice = time_slice(reference_tx)
    rel_ent_inst_edge = uzr | in_rel[BT.RAE_INSTANCE_EDGE]
    all_edges = [(e, "low_lvl", "") for e in rel_ent_inst_edge | in_rels[BT] | collect]
    add_directed_rt_to_list(uzr | in_rels[RT] | filter[lambda z: BT(z) != BT.RAE_INSTANCE_EDGE] | collect, "in")
    add_directed_rt_to_list(uzr | out_rels[RT] | collect, "out")
    all_edges.sort(key=edge_time_by_type)
    descriptions_and_txs = []
    termination_edge_description = [""]
    # Loop over all edges and filter based on edge_type and on the reference_tx value
    for e in all_edges:
        edge, edge_type, edge_state = e
        if edge_type == "low_lvl" and edge | source | time_slice | collect <= reference_tx_timeslice:
            edge_description = (f"{fill_str_to_length(generate_low_level_description(edge), 65)}  "
                                f"[{readable_datetime_from_tx_uzr(edge | source | collect)}]", edge)
            if BT(edge) == BT.TERMINATION_EDGE:
                termination_edge_description = edge_description
            else:
                descriptions_and_txs.append(edge_description)
        else:
            if (edge_state == "Instantiated" and time_slice(edge | instantiation_tx | collect) <= reference_tx_timeslice) or\
                    (edge_state == "Terminated" and time_slice(edge | termination_tx | collect) <= reference_tx_timeslice):
                descriptions_and_txs.append(
                    (f"{fill_str_to_length(generate_rt_description(edge, edge_type, edge_state), 54)}", edge))

    filler = '\n    |\n    |\n'
    return f"""\
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ Timeline ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

{termination_edge_description[0]}
    ^
   ^^^
    |
{filler.join([x[0] for x in descriptions_and_txs[1:][::-1]])}
    |
    |
{descriptions_and_txs[0][0]}
-----------------------------------------------------------------------------------------------------------------\n
"""


def graph_info(g) -> str:
    bl = blobs(g)

    def simple_lengths(g) -> str:
        return (f"{g.graph_data.write_head} blobs, "
                + str(length(bl | filter[BT.TX_EVENT_NODE])) + " TXs, "
                + str(length(bl | filter[BT.ENTITY_NODE])) + " ETs, "
                + str(length(bl | filter[BT.ATOMIC_ENTITY_NODE])) + " AETs, "
                + str(length(bl | filter[BT.RELATION_EDGE])) + " RTs"
                )

    return f"""
======================================================================================================================
=============================================== Graph {uid(g)} ===============================================
===================================== Seen from: {str(now())} =========================================
======================================================================================================================

instantiation:          {readable_datetime_from_tx_uzr(g | All[TX] | first | collect) if length(g | All[TX] | collect) > 0 else "NA"}
last change:            {readable_datetime_from_tx_uzr(g | All[TX] | last | collect)  if len(g | All[TX] | collect) > 0 else "NA"}
current tags:           {g.graph_data.tag_list}
summary:                {simple_lengths(g)}
size:                   {round((g.graph_data.write_head * 16) / 1E6, 3)}MB

^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ Atomic Entities ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
{type_summary_view(bl, g, BT.ATOMIC_ENTITY_NODE)}
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ Entities ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
{type_summary_view(bl, g, BT.ENTITY_NODE)}
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^  Relations ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
{type_summary_view(bl, g, BT.RELATION_EDGE)}
"""


def type_summary_view(bl: EZefRefs, g: Graph, bt_filter: BlobType) -> str:
    def aet_et_rt_string_view(summary: Tuple[int, int, tuple]) -> str:
        aet_or_et, total, alive = summary[0], summary[1], summary[2]
        return f"[{fill_str_to_length(str(f'{total} total, {alive} alive] '), 30)}{aet_or_et}\n"

    def triple_string_view(summary: Tuple[int, int, tuple]) -> str:
        triple, total, alive = summary[0], summary[1], summary[2]
        return (f"       [{fill_str_to_length(str(f'{total} total, {alive} alive] '), 30)}"
                f"({triple[0]}, {triple[1]}, {triple[2]})\n")

    def find_alive_count_of_triple(triple: Tuple[str]) -> int:
        # TODO Replace this when triple lookup is implemented for instances zefop
        # return seq([z for z in
        #     (g[internals.root_node_blob_index()] | out_rels[RT] | filter[BT.TO_DELEGATE_EDGE] | target | filter[RT(triple[1][3:])]
        #       | only | out_rels[RT] | filter[lambda z: not internals.is_delegate(z)] | target
        #       | filter[lambda x: zr_type(x | source | collect) == triple[0] and zr_type(x | target | collect) == triple[2]]
        #             | collect)])\
        #     .sum(lambda x: len(x | now | all | collect))
        temp = g[internals.root_node_blob_index()] | out_rels[BT] | filter[BT.TO_DELEGATE_EDGE] | target | filter[RT(triple[1][3:])] | single # | out_rels[RT] #| filter[lambda z: not internals.is_delegate(z)] | target
        temp = temp | out_rels[BT] | filter[lambda z: not internals.is_delegate(z)] | target
        temp = temp | filter[lambda x: zr_type(x | source | collect) == triple[0] and zr_type(x | target | collect) == triple[2]]
        temp = temp | collect
        return sum(len(x | now | all | collect) for x in temp)

    def find_triples_of_rt(rt: str) -> str:
        return "".join(
            seq([(zr_type(z | source | collect), zr_type(z), zr_type(z | target | collect)) for z in
                bl | filter[RT(rt)] | filter[lambda z: not internals.is_delegate(z)]])
            .group_by(lambda x: x)
            .map(lambda x: (x[0], len(x[1]), find_alive_count_of_triple(x[0])))
            .map(triple_string_view)
        )
    # TODO find correct alive instances for RT
    return "".join(
        seq([z for z in bl | filter[bt_filter] | filter[lambda z: not internals.is_delegate(z)]])
        .group_by(lambda z: zr_type(z))
        .map(lambda x: (x[0], len(x[1]), len(x[1][0] | delegate_of | now | all | collect)))
        .map(lambda x: aet_et_rt_string_view(x) + find_triples_of_rt(
            x[0][3:]) if bt_filter == BT.RELATION_EDGE else aet_et_rt_string_view(x))
    )


# +\
# indent_lines(relations_view(zr_or_uzr)) + "\n" +\
# indent_lines(timeline_view(zr_or_uzr))


# def src_dst_info(uzr) -> str:
#     return f"""
#     ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ Source and Target ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

#     ({uzr | source | uid}: {zr_type(uzr | source)})-----(z)----->({uzr | target | uid}: {zr_type(uzr | target)})
# """



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ctypes import *


def convert_from_v1(blobs):
    out = {}

    for index,item in blobs.items():
        new_item = {**item}
        new_item["_old_index"] = index
        if "uid" in item:
            new_item["_internalUID"] = item["uid"]
        else:
            new_item["_internalUID"] = index

        out[index] = new_item

    return out



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import *
from ...ops import *
from .server import start_server

from zef.core.logger import log

import argparse
parser = argparse.ArgumentParser()
parser.add_argument("schema_file", type=str, metavar="SCHEMA-FILE",
                    help="The graphql schema file.")
parser.add_argument("data_tag", type=str, metavar="DATA-TAG",
                    help="The tag of the Zef graph with the data (not the schema).")
parser.add_argument("--port", type=int, default=443)
parser.add_argument("--bind", type=str, default="0.0.0.0",
                    help="IP address to bind to.")
parser.add_argument("--no-host-role", action="store_false", dest="host_role",
                    help="Whether the server should acquire host role on the data graph.")
parser.add_argument("--create", action="store_true",
                    help="If the data graph does not exist, then create a new blank graph.")
args = parser.parse_args()

from .schema_file_parser import parse_partial_graphql, json_to_minimal_nodes

g_schema = Graph()
schema_gql = args.schema_file | read_file | run | get["content"] | collect
parsed_dict = parse_partial_graphql(schema_gql)
commands = json_to_minimal_nodes(parsed_dict)
r = commands | transact[g_schema] | run
root = r["root"]
log.info(f"Created schema graph from '{args.schema_file}'")

guid = lookup_uid(args.data_tag)
if guid is None:
    if not args.create:
        print("Graph is not known to us, and create is False. Exiting.")
        raise SystemExit(2)
    g_data = Graph(True)
    try:
        g_data | tag[args.data_tag] | run
    except:
        print("Unable to create and tag graph with '{args.data_tag}'. Maybe this tag is already taken by another user?")
        raise SystemExit(2)
    log.info("Created data graph with tag", tag=args.data_tag)
else:
    g_data = Graph(args.data_tag)
    log.info("Loaded existing data graph")

if args.host_role:
    make_primary(g_data)
    log.info("Obtained host role on data graph")

server_uuid = start_server(root, g_data, args.port, args.bind)

import time
try:
    while True:
        time.sleep(1)
except KeyboardInterrupt:
    pass
finally:
    Effect({
        "type": FX.HTTP.StopServer,
        "server_uuid": server_uuid
    }) | run


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest  # pytest takes ages to run anything as soon as anything from zef is imported
from zef import *
from zef.ops import *


class MyTestCase(unittest.TestCase):
    def test_serialization(self):
        g = Graph()
        z = ET.Machine | g | run
        a,b,c = (z, RT.Something, "data") | g | run

        to_check = [z,a,b,c]
        to_check += [origin_rae(a)]
        to_check += [origin_rae(b)]
        to_check += [origin_rae(c)]
        to_check += [uid(z)]

        to_check += [
            {"key": 5,
             z: ([z,z,(z,z),[z,z], {z: z}], z)
             }
        ]

        for item in to_check:
            self.assertEqual(deserialize(serialize(item)), item)

if __name__ == '__main__':
    unittest.main()



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

import pygame
import OpenGL.GL as gl

from imgui.integrations.pygame import PygameRenderer
import imgui

def main(window_name, init_state, render_func, *args, **kwds):
    pygame.init()
    size = 800, 600

    pygame.display.set_mode(size, pygame.DOUBLEBUF | pygame.OPENGL | pygame.RESIZABLE)

    imgui.create_context()
    impl = PygameRenderer()

    io = imgui.get_io()
    io.display_size = size

    state = init_state(*args, **kwds)

    while 1:
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                sys.exit()

            impl.process_event(event)

        imgui.new_frame()

        render_func(state)

        # note: cannot use screen.fill((1, 1, 1)) because pygame's screen
        #       does not support fill() on OpenGL sufraces
        gl.glClearColor(1, 1, 1, 1)
        gl.glClear(gl.GL_COLOR_BUFFER_BIT)
        imgui.render()
        impl.render(imgui.get_draw_data())

        pygame.display.flip()


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .value_type import *


# the following functions add a layer of indirection to prevent circular imports

def graph_ctor(*args, **kwargs):
    from ... import core
    return core.Graph(*args, **kwargs)



def time_ctor(*args, **kwargs):
    from ... import pyzef
    return pyzef.main.Time(*args, **kwargs)


def bytes_ctor(*args, **kwargs):
    # from ... import core
    from ...core.bytes import Bytes_
    # return core.bytes.Bytes_(*args, **kwargs)
    return Bytes_(*args, **kwargs)


def decimal_ctor(*args, **kwargs):
    # from ... import core
    from ...core.decimal import Decimal_
    # return core.bytes.Bytes_(*args, **kwargs)
    return Decimal_(*args, **kwargs)

# def error_ctor(*args, **kwargs):
#     from ... import core
#     return core.error._Error.Error(*args, **kwargs)



Nil        = ValueType_(type_name='Nil',        constructor_func=None)
Any        = ValueType_(type_name='Any',        constructor_func=None)
Bool       = ValueType_(type_name='Bool',       constructor_func=bool)
Int        = ValueType_(type_name='Int',        constructor_func=int)
Float      = ValueType_(type_name='Float',      constructor_func=float)
String     = ValueType_(type_name='String',     constructor_func=str)
Bytes      = ValueType_(type_name='Bytes',      constructor_func=bytes_ctor)
Decimal    = ValueType_(type_name='Decimal',    constructor_func=decimal_ctor)
List       = ValueType_(type_name='List',       constructor_func=tuple)
Dict       = ValueType_(type_name='Dict',       constructor_func=dict)
Set        = ValueType_(type_name='Set',        constructor_func=set)
EZefRef    = ValueType_(type_name='EZefRef',    constructor_func=None)    
ZefRef     = ValueType_(type_name='ZefRef',     constructor_func=None)     
Graph      = ValueType_(type_name='Graph',      constructor_func=graph_ctor)     
GraphSlice = ValueType_(type_name='GraphSlice', constructor_func=None)     
FlatGraph  = ValueType_(type_name='FlatGraph',  constructor_func=None)     
ZefOp      = ValueType_(type_name='ZefOp',      constructor_func=None)     
Stream     = ValueType_(type_name='Stream',     constructor_func=None)     
TX         = ValueType_(type_name='TX',         constructor_func=None)     
Time       = ValueType_(type_name='Time',       constructor_func=time_ctor)
Error      = ValueType_(type_name='Error',      constructor_func=None)
Image      = ValueType_(type_name='Image',      constructor_func=None)
ValueType  = ValueType_(type_name='ValueType',  constructor_func=ValueType_)


# QuantityInt= ValueType_(type_name='QuantityInt',constructor_func=None)
# QuantityFloat= ValueType_(type_name='QuantityFloat',constructor_func=None)

UID        = ValueType_(type_name='UID',        constructor_func=None)       
BaseUID    = ValueType_(type_name='BaseUID',    constructor_func=None)       
ZefRefUID  = ValueType_(type_name='ZefRefUID',  constructor_func=None)        
EternalUID = ValueType_(type_name='EternalUID', constructor_func=None)       

Instantiated = ValueType_(type_name='Instantiated', constructor_func=None)       
Terminated = ValueType_(type_name='Terminated', constructor_func=None)       
Assigned   = ValueType_(type_name='Assigned',   constructor_func=None)       
Tagged     = ValueType_(type_name='Tagged',     constructor_func=None)       

LazyValue  = ValueType_(type_name='LazyValue',  constructor_func=None) 
Awaitable  = ValueType_(type_name='Awaitable',  constructor_func=None) 



AET        = ValueType_(type_name='AET',        constructor_func=None)
ET         = ValueType_(type_name='ET',         constructor_func=None)
RT         = ValueType_(type_name='RT',         constructor_func=None)
BT         = ValueType_(type_name='BT',         constructor_func=None)
Enum       = ValueType_(type_name='Enum',       constructor_func=None)
Record     = ValueType_(type_name='Record',     constructor_func=None)       
Function   = ValueType_(type_name='Function',   constructor_func=None)
GraphDelta = ValueType_(type_name='GraphDelta', constructor_func=None)     
Query      = ValueType_(type_name='Query',      constructor_func=None) 
DeltaQuery = ValueType_(type_name='DeltaQuery', constructor_func=None) 
Effect     = ValueType_(type_name='Effect',     constructor_func=None) 
DataFrame  = ValueType_(type_name='DataFrame', constructor_func=None)     
EZefRefs   = ValueType_(type_name='EZefRefs', constructor_func=None)     
EZefRefss  = ValueType_(type_name='EZefRefss', constructor_func=None)     
ZefRefs    = ValueType_(type_name='ZefRefs', constructor_func=None)     
ZefRefss   = ValueType_(type_name='ZefRefss', constructor_func=None)     



# These are special classes: using them with [...] returns a ValueType_ though
Union = UnionClass()
Intersection = IntersectionClass()
SetOf = SetOfClass()



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .._ops import *
from .._core import *

##############################
# * Call
#----------------------------

# Call implements the laziness part of the op chain. It's use on the user-facing
# side is discouraged, although zef functions will be converted to call ops.

def call_implementation(input, f, *args):
    from typing import Generator, Iterable, Iterator
    if isinstance(input, Generator) or isinstance(input, Iterator):
        input = [i for i in input]
    return f(input, *args)

def call_type_info(op, curr_type):
    # TODO
    return VT.Any

##############################
# * Unpack
#----------------------------

def unpack_implementation(inputs, f):
    if isinstance(inputs, tuple) or isinstance(inputs, list):
        return f(*inputs)
    elif isinstance(inputs, dict):
        # if this has wrapped a call[...] then we need to unpack the function
        # inside of it, as call itself does not handle kwargs

        # Why is this nesting so confusing???
        while isinstance(f, ZefOp):
            assert len(f.el_ops) == 1 and f.el_ops[0][0] == RT.Call, "Can only support a single call[] zefop when using dictionary unpack"
            assert len(f.el_ops[0][1]) == 1, "Can't support curried arguments yet, e.g. unpack[func[something][1][2]], when using dicitonary unpack"
            f = f.el_ops[0][1][0]
        return f(**inputs)
    else:
        raise Exception("unpack requires either a tuple or a dictionary as input")

def unpack_type_info(op, curr_type):
    # TODO
    return VT.Any





# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Heavily modified from https://stackoverflow.com/a/2406043

import types
# Getting really frustrated with circular imports
beingimported = []
originalimport = __import__
def newimport(modulename, globs=None, locals=None, fromlist=(), level=0):
    if isinstance(globs, dict):
        parent_name = globs.get("__name__", None)
        is_package = (parent_name == globs.get("__package__", "1"))
    else:
        parent_name = ""
        is_package = False
    if level == 0:
        qual_name = modulename
    else:
        rel_level = level
        if is_package:
            rel_level -= 1
        if rel_level == 0:
            rel_name = parent_name
        else:
            rel_name = '.'.join(parent_name.split('.')[:-rel_level])
        if modulename == "":
            qual_name = rel_name
        else:
            qual_name = rel_name + '.' + modulename

    # Trying method with all fromlist used instead
    was_already_present = qual_name in beingimported
    beingimported.append(qual_name)

    sub_check = {}
    if isinstance(fromlist, tuple):
        for item in fromlist:
            if item == "*":
                continue
            sub_qual_name = qual_name + "." + item
            sub_check[item] = (sub_qual_name, sub_qual_name in beingimported)
            beingimported.append(sub_qual_name)

    try:
        result = originalimport(modulename, globs, locals, fromlist, level)

        try:
            if "zef" in parent_name:
                # only_importing_submodules = isinstance(fromlist, tuple) and ('*' not in fromlist) and all(isinstance(getattr(result, x), types.ModuleType) for x in fromlist)
                # except Exception as exc:
                #     print(f"Errored: {exc}")
                #     print(dir(result))
                #     only_importing_submodules = False
                only_importing_submodules = (len(sub_check) > 0)
                for sub_name, (sub_qual_name, sub_was_previous) in sub_check.items():
                    try:
                        sub_obj = getattr(result, sub_name)
                    except:
                        print(f"!!! After trying to import item {sub_qual_name=} it wasn't in the module!")
                        raise
                    if isinstance(sub_obj, types.ModuleType):
                        if sub_was_previous:
                            import traceback
                            frame = traceback.extract_stack()[-2]
                            print("=========")
                            print(f"Importing in circles {sub_qual_name=} from import {qual_name=} {level=} {parent_name=} {is_package=} {fromlist=}")
                            print("    Currently importing -> ", beingimported)
                            print_all_imports_in_stack()
                    else:
                        only_importing_submodules = False
                if was_already_present and not only_importing_submodules:
                    # We don't worry about importing submodules from us when we're a package
                    is_submodule_import = (is_package and level == 1)
                    if not is_submodule_import:
                        import traceback
                        frame = traceback.extract_stack()[-2]
                        print("=========")
                        print(f"Importing in circles {modulename=} {qual_name=} {level=} {parent_name=} {is_package=} {fromlist=}")
                        print("    Currently importing -> ", beingimported)
                        print_all_imports_in_stack()
        except Exception as exc:
            print("Ignoring exc in detecting circular imports")
    finally:
        def remove_last_occurrence(l, x):
            if x in l:
                l.reverse()
                l.remove(x)
                l.reverse()
        remove_last_occurrence(beingimported, qual_name)
        for (sub_qual_name, sub_was_previous) in sub_check.values():
            remove_last_occurrence(beingimported, sub_qual_name)

    return result

def check_circular_imports():
    import builtins
    builtins.__import__ = newimport

def disable_check_circular_imports():
    import builtins
    builtins.__import__ = originalimport



def print_all_imports_in_stack():
    import traceback
    stack = traceback.extract_stack()[:-2]
    stack.reverse()

    for frame in stack:
        if "import" in frame.line and 'circular_imports' not in frame.filename:
            print(f"   {frame.filename}: {frame.lineno} -- {frame.line}")
            
    


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .fx_types import Effect


def graph_transaction_handler(eff: Effect):
    """[summary]

    Args:
        eff (Effect): Effect({
            "type": FX.TX.Transact,
            "target_graph": g1,
            "commands": list_of_commands
        })
    """

    from ..graph_delta import perform_transaction_commands, filter_temporary_ids, unpack_receipt

    if eff.d["target_graph"].graph_data.is_primary_instance:
        receipt = perform_transaction_commands(eff.d['commands'], eff.d['target_graph'])
    else:
        from ..overrides import merge
        receipt = merge(eff.d["commands"], eff.d["target_graph"])
    
    # we need to forward this if it is in the effect
    if 'unpacking_template' in eff.d:
        return unpack_receipt(eff.d['unpacking_template'], receipt)

    receipt = filter_temporary_ids(receipt)
    return receipt


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This file contains tests for things that were bugs at one point. Does not need
# to be organised at all

import unittest  # pytest takes ages to run anything as soon as anything from zef is imported
import os
from zef import *
from zef.ops import *
import zef


class MyTestCase(unittest.TestCase):

    def test_linking_tx(self):
        g = Graph()

        # Get the first tx on the graph
        tx_node = g | all[TX] | first | collect

        r = [
            (tx_node, RT.User, "someone"),
        ] | transact[g] | run

        self.assertEqual("someone", tx_node | now | Out[RT.User] | value | collect)

        r = [
            (tx_node, RT.Something, g | root | collect),
        ] | transact[g] | run

        # Note: the root has delegates coming off of it, so we need to filter by only the concrete relation.
        self.assertEqual(g | now | root | in_rels[RT.Something] | filter[Not[is_a[Delegate]]] | single | collect,
                         tx_node | now | out_rel[RT.Something] | collect)

    def test_linking_with_transaction(self):
        g = Graph()

        with Transaction(g) as ctx:
            (ctx, RT.User, "Someone") | g | run

        self.assertEqual(g | all[TX] | last | now | Out[RT.User] | value | collect,
                         "Someone")

    def test_attaching_to_delegates(self):
        g = Graph()

        with Transaction(g) as ctx:
            (ctx, RT.User, "Someone") | g | run

        (delegate_of(ctx), RT.Fixed, True) | g | run
        # Note: True so that we create the delegate
        (g | root | delegate_of[True] | collect, RT.Root, True) | g | run
        # Note: trying alternative with abstract delegate
        (g | root | to_delegate | delegate_of | collect, RT.Second, True) | g | run

    def test_failure_on_other_graph(self):
        g = Graph()
        g2 = Graph()

        tx_node = g | all[TX] | first | collect
        with self.assertRaisesRegex(Exception, "Can only merge TXNode and Root onto the same graph"):
            (tx_node, RT.User, "someone") | g2 | run

if __name__ == '__main__':
    unittest.main()



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import zstd

class Image:
    def __init__(self, data, format='svg'):
        self.format = format
        self.compression = 'zstd'
        self.buffer = zstd.compress(data)
    def _repr_svg_(self):
        # this function must return a str
        return zstd.decompress(self.buffer).decode("utf-8") if self.format=='svg' else None
    def _repr_png_(self):
        # this function must return bytes
        return zstd.decompress(self.buffer) if self.format=='png' else None
    def _repr_jpeg_(self):
        # this function must return bytes
        return zstd.decompress(self.buffer) if self.format=='jpg' else None

    def _view(self, format=None):
        if format is None:
            format = self.format

        import tempfile, os
        fd,filename = tempfile.mkstemp(f".{format}")
        file = os.fdopen(fd, 'wb')
        try:
            # TODO: In the future, do a conversion of file format
            if format == "svg":
                data = self._repr_svg_()
            elif format == "png":
                data = self._repr_png_()
            elif format == "jpeg":
                data = self._repr_jpeg_()

            if data is None:
                raise Exception(f"Can't convert '{self.format}' to '{format}'")
                
            if isinstance(data, str):
                data = data.encode("utf-8")
            file.write(data)
        finally:
            file.close()

        from . import FX, Effect
        from ._ops import run
        Effect({
            "type": FX.LocalFile.SystemOpenWith,
            "filepath": filename
        }) | run


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .. import *
from .._ops import *

def graphviz_imp(zz, *flags):
    """
    Returns a GraphViz visualization of the graph passed in.
    By default it returns an error if more than 1000 RAEs are to be shown.

    Zef graphs are meta-graphs: relations can themselves have outgoing relations
    which allows mimicing the behavior of property graphs (with nodes/relations 
    allowed to contain "objects" / dictionaries).

    GraphViz does not allow plotting relations out of relations (please let us
    know if we're wrong here and you know a way!). Therefore we need to 
    map the meta-graph onto a larger true graph (each relation becoming a node)
    and subsequently plot this. The downside is the visual clarity:
    It may be hard to see which edges/lines are the true edges and which lines
    are other incoming/outgoing edges to the actual edge.
    
    As a slight improvement, each relation's color is based on it's type: this 
    allows identifying the actual relation as the one with and incoming and
    outgoing arrow of the same color in case of ambiguity.

    ---- Examples ----
    >>> g | now | all | graph_viz | collect
    >>> g | now | all | graph_viz[simplified] | collect     # No extra nodes introduced for relations. Hence: can't show relations out of relations.
    >>> g | all | graph_viz | collect                       # show the full eternal graph


    ---- Signature ----
    List[EZefRef] -> Image
    List[ZefRef] -> Image
    (List[ZefRef], ZefOp) -> Image
    """
    if isinstance(zz, FlatGraph): 
        g = Graph()
        zz | transact[g] | run 
        zz = g | now | all[RAE] | collect
    from functools import lru_cache
    try:
        import graphviz
    except:
        msg = """ 
        Error importing the Python 'GraphViz' package.
        Two steps are required:
            1) GraphViz on the system needs to be installed:
                e.g. on Ubuntu 'sudo apt install graphviz'
                e.g. on Mac 'brew install graphviz'
                google otherwise :)
            2) Python package: e.g. 'pip3 install graphviz'
        """
        raise RuntimeError(msg)

    def make_color_generator():
        """
        returns a generator that first picks from a 
        predefined list of colors and switches to generating
        random colors once the list exhausts.
        """
        import random
        color_list = (
            '#115A63',
            '#78C25C',
            '#6f42b3',
            '#2fb6fc',
            '#3e4337',
            '#31BBA5',
            '#4342d5',
            '#5edfcb',
            '#078f01',
            '#2c3167',
            '#402667',
            '#515252',
            '#604862',
        )
        yield from color_list        
        alphabet="0123456789ABCDEF"
        while True:
            r = random.SystemRandom()
            yield '#' + ''.join([r.choice(alphabet) for i in range(6)])
    
    
    def make_color_generator_edges():
        """
        returns a generator that first picks from a 
        predefined list of colors and switches to generating
        random colors once the list exhausts.
        """
        import random
        color_list = (
            '#00171f',
            '#777777',
            '#99374F',
            '#124763',
            '#402667',
            '#604862',
        )
        yield from color_list        
        alphabet="0123456789ABCDEF"
        while True:
            r = random.SystemRandom()
            yield '#' + ''.join([r.choice(alphabet) for i in range(6)])

    color_gen = make_color_generator()
    color_gen_edges = make_color_generator_edges()

    @lru_cache(maxsize=10000)
    def hex_color_for_type(some_type):
        return next(color_gen)      

    @lru_cache(maxsize=10000)
    def hex_color_for_type_edges(some_type):
        return next(color_gen_edges)       
    
    graph_ctor_kwargs = ({} 
        | if_then_else_apply[lambda _: KW.neato in flags][insert['engine']['neato']][identity]
        | collect
    )


    G = graphviz.Digraph('G', **graph_ctor_kwargs)
    if isinstance(zz, Graph): zz = all(zz)
    zz = list(zz)
    if len(zz) == 0:
        return G            # exit early if there is nothing to plot
    if isinstance(zz, ZefRefs) or (isinstance(zz, list) and isinstance(zz[0], ZefRef)):
        plotting_eternal_graph = False
    else:
        plotting_eternal_graph = True

    edge_colors = {}
    def nice_name(z: EZefRef) -> str:
        if BT(z)==BT.ENTITY_NODE:
            return f"ET.{str(ET(z))}"        
        if BT(z)==BT.TX_EVENT_NODE:
            return f"TX"
        if BT(z)==BT.ATOMIC_ENTITY_NODE:
            def val_to_str(z):
                v = value(z)
                if v is None: return 'None'
                elif isinstance(v, str): return f'"{v}"' if len(v) < 20 else f'"{v[:18]}..."'
                else: return str(v)
            val_maybe = '' if isinstance(z, EZefRef) else (f"\n▷{val_to_str(z)}")
            return f"AET.{AET(z)}{val_maybe}"        
        if BT(z)==BT.RELATION_EDGE:
            return f"RT.{str(RT(z))}"        
        else:
            return f"BT.{str(BT(z))}"[:-5]
    
    def nice_color(z: EZefRef)->str:
        if BT(z) == BT.ENTITY_NODE:
            return hex_color_for_type(ET(z))
        if BT(z) == BT.ATOMIC_ENTITY_NODE:
            return hex_color_for_type(AET(z))
            
        if BT(z) in {BT.TX_EVENT_NODE, BT.NEXT_TX_EDGE}:
            return "#991155"
        if BT(z) == BT.RELATION_EDGE:
            return "#000000" if plotting_eternal_graph else hex_color_for_type_edges(RT(z))
        else:            
            return hex_color_for_type(BT(z))

    def nice_shape(z):
        if BT(z) == BT.TX_EVENT_NODE:
            return 'triangle'
        elif BT(z) == BT.ENTITY_NODE:
            return 'oval'            
        elif BT(z) == BT.ATOMIC_ENTITY_NODE:
            return 'rectangle'
        else:
            return 'diamond'

    def nice_style(z):
        if BT(z) in {BT.ENTITY_NODE, BT.ATOMIC_ENTITY_NODE}:
            return '' if internals.is_delegate(z) else 'filled'
        else:
            return ''
    
    def is_node(z):
        return str(z | BT | collect)[-5:] == '_NODE'
    
    def add_node(z):
        G.node(str(index(z)),  label=nice_name(z), fontsize='12px', color=nice_color(z), shape=nice_shape(z), style=nice_style(z))
        
    def add_edge_as_gv_node(z):
        col = nice_color(z)
        edge_colors[index(z)] = col
        G.node(str(index(z)), label=nice_name(z), color=col, shape='none', style='', fontsize='10px', fontcolor=col)
        
    def add_gv_edges(z):
        # ------- hacky / dirty way for now: just try going to source / target and swallow error if anything fails :| --------
        try:
            col = edge_colors[index(z)]   # this was determined above
            G.edge(str(index(z | source | collect)), str(index(z)), color=col, dir="back", arrowtail="dot")
            G.edge(str(index(z)), str(index(z | target | collect)), color=col, dir="forward", arrowhead="normal")
        except Exception as e:
            #print(f"add_gv_edges failed for {z} {e=}")
            pass
        
    nodes, edges = zz | group_by[is_node][(True,False)] | map[second] |  collect
    nodes | for_each[add_node]
    edges | for_each[add_edge_as_gv_node]
    edges | for_each[add_gv_edges]

    # Convert the Graphviz Graph to svg string and then to an Image
    return Image(G.pipe(format='svg'))


def graphviz_tp(x):
    return VT.Any           # return a VT.Image in future?



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .._core import *
from .._ops import *
from .fx_types import _Effect_Class

from ...pyzef.admins import add_right, remove_right

def main_implementation(privilege, target, action, user):
    assert privilege in [KW.view, KW.append, KW.host, KW.discover, KW.modify_rights]

    s_privilege = "Allow" + to_pascal_case(str(privilege))

    if is_a(target, Graph):
        s_target = f"graph:{base_uid(target)}"
    else:
        raise NotImplementedError("TODO: handle targets other than graphs")

    assert isinstance(user, str), "Users need to be strings at the moment"
    # This is to allow groups as well as users but is only temrporary while we're using strings.
    if ':' not in user:
        s_subject = "user:" + user
    else:
        s_subject = user

    ret = {}
    try:
        if action == KW.grant:
            add_right(s_subject, s_target, s_privilege)
        elif action == KW.revoke:
            remove_right(s_subject, s_target, s_privilege)
        else:
            raise Exception(f"Unknown action {action}")
        ret["action_taken"] = True

    except RuntimeError as exc:
        ret["extra_info"] = str(exc)
        if "Right not present to be removed" in str(exc):
            ret["action_taken"] = False
        elif "Right already present" in str(exc):
            ret["action_taken"] = False
        else:
            raise

    return ret

def privileges_grantview_handler(eff: _Effect_Class):
    return main_implementation(KW.view, eff.d["target"], KW.grant, eff.d["user"])

def privileges_grantappend_handler(eff: _Effect_Class):
    return main_implementation(KW.append, eff.d["target"], KW.grant, eff.d["user"])

def privileges_granthost_handler(eff: _Effect_Class):
    return main_implementation(KW.host, eff.d["target"], KW.grant, eff.d["user"])

def privileges_grantdiscover_handler(eff: _Effect_Class):
    return main_implementation(KW.discover, eff.d["target"], KW.grant, eff.d["user"])

def privileges_grantmodifyrights_handler(eff: _Effect_Class):
    return main_implementation(KW.modify_rights, eff.d["target"], KW.grant, eff.d["user"])

def privileges_revokeview_handler(eff: _Effect_Class):
    return main_implementation(KW.view, eff.d["target"], KW.revoke, eff.d["user"])

def privileges_revokeappend_handler(eff: _Effect_Class):
    return main_implementation(KW.append, eff.d["target"], KW.revoke, eff.d["user"])

def privileges_revokehost_handler(eff: _Effect_Class):
    return main_implementation(KW.host, eff.d["target"], KW.revoke, eff.d["user"])

def privileges_revokediscover_handler(eff: _Effect_Class):
    return main_implementation(KW.discover, eff.d["target"], KW.revoke, eff.d["user"])

def privileges_revokemodifyrights_handler(eff: _Effect_Class):
    return main_implementation(KW.modify_rights, eff.d["target"], KW.revoke, eff.d["user"])


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Note: This should never depend on core at the top level!

from ...pyzef.internals import (
    AtomicEntityTypeStruct,
    BaseUID,
    BlobType,
    BlobTypeStruct,
    EntityTypeStruct,
    EternalUID,
    FinishTransaction,
    GraphData,
    Graph_from_ptr,
    KW,
    MMAP_STYLE_ANONYMOUS,
    MMAP_STYLE_AUTO,
    MMAP_STYLE_FILE_BACKED,
    NamedHeadRange,
    ObservablesDictElement,
    RelationTypeStruct,
    StartTransactionReturnTx,
    Subscription,
    UpdateHeads,
    UpdatePayload,
    ZefEnumStruct,
    ZefEnumStructPartial,
    ZefRefUID,
    add_entity_type,
    add_enum_type,
    add_relation_type,
    add_keyword,
    all_entity_types,
    all_enum_types_and_values,
    all_relation_types,
    apply_update,
    blob_to_json,
    create_graph_from_bytes,
    create_update_heads,
    create_update_payload,
    created_token_list,
    current_zefdb_protocol_version,
    delete_graphdata,
    delegate_to_ezr,
    early_token_list,
    ezr_to_delegate,
    get_data_layout_version_info,
    get_enum_value_from_string,
    get_graph_revision_info,
    get_latest_complete_tx_node,
    get_local_process_graph,
    graph_as_UpdatePayload,
    gtd_info_str,
    has_delegate,
    heads_apply,
    initialise_butler,
    initialise_butler_as_master,
    is_BaseUID,
    is_EternalUID,
    is_ZefRefUID,
    is_aet_a_enum,
    is_aet_a_quantity_float,
    is_aet_a_quantity_int,
    is_any_UID,
    is_delegate,
    is_delegate_group,
    is_root,
    is_up_to_date,
    list_graph_manager_uids,
    login,
    logout,
    make_random_uid,
    max_zefdb_protocol_version,
    memory_details,
    merge_atomic_entity_,
    merge_entity_,
    merge_relation_,
    num_blob_indexes_to_move,
    pageout,
    parse_payload_update_heads,
    partial_hash,
    root_node_blob_index,
    set_data_layout_version_info,
    set_graph_revision_info,
    show_blob_details,
    size_of_blob,
    sort_uzr,
    sort_uzr_magnitude,
    sort_zr,
    sort_zr_magnitude,
    start_connection,
    stop_butler,
    stop_connection,
    to_uid,
    validate_message_version,
    wait_for_auth,
)

from ...pyzef.main import ZefRef

# ----------------------------- monkey patch to allow dispatching on enums ---------------------------------------
# For some reason we need to pass it through a lambda fct to get the bindings to work.
# BlobType.__rrshift__ = lambda bt, any_zr_type: fct__rrshift__(bt, any_zr_type)
# BlobType.__rlshift__ = lambda bt, any_zr_type: fct__rlshift__(bt, any_zr_type)
# BlobType.__gt__ = lambda bt, any_zr_type: fct__gt__(bt, any_zr_type)
# BlobType.__lt__ = lambda bt, any_zr_type: fct__lt__(bt, any_zr_type)

BlobType.__str__ = lambda self: repr(self).split(':')[0][10:]   # strip away the 'BlobType. ...' and out <>

from .rel_ent_classes import ET, RT, EN, AET
from .merges import register_merge_handler
register_merge_handler()

BT = BlobTypeStruct()

def safe_current_task():
    import asyncio
    try:
        return asyncio.current_task()
    except RuntimeError:
        return None
    
global_transaction_task = {}
from contextlib import contextmanager
@contextmanager
def Transaction(g, wait=None):
    global global_transaction_task
    from ...pyzef.zefops import uid

    cur_task = safe_current_task()
    prev_val = global_transaction_task.get(uid(g), None)

    if prev_val is not None and prev_val != cur_task:
        raise Exception("Can't open a Transaction from a different task to the original Transaction. Note that ZefDB is not safe to create simultaneous transactions from different asyncio tasks. Ideally, there should be no asyncio task switch occuring during a transaction.")

    global_transaction_task[uid(g)] = cur_task
    current_tx = StartTransactionReturnTx(g)
    try:
        yield current_tx
    finally:
        if prev_val is None and uid(g) in global_transaction_task:
            del global_transaction_task[uid(g)]
        else:
            global_transaction_task[uid(g)] = prev_val
        if wait is None:
            FinishTransaction(g)
        else:
            FinishTransaction(g, wait)



def assign_value_imp(z, value):
    from .._ops import is_a
    from ...pyzef.zefops import SerializedValue, assign_value as c_assign_value
    from ..serialization import serialize, serialization_mapping

    assert isinstance(z, ZefRef) or isinstance(z, EZefRef)
    if is_a(z, AET.Serialized):
        if type(z) in serialization_mapping:
            from json import dumps
            value = SerializedValue("tools.serialize", dumps(serialize(value)))
        else:
            raise Exception(f"Don't know how to serialize a type of {val}")
    c_assign_value(z, value)



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    "update_items_from_json", 
    "update_items_convert_value",
    "AnyT",
    "zef_spec_check_relent",
    "zef_spec_check_relent_details",
    "eq",
    "noteq",
    "zef_spec_check_relation_details",
    "zef_spec_check_structure",
    "all_connected_entities",
    "cprint",
    "inject",
    "collect_subgraph",
    "timeit",
]

from ...core import *
from ...ops import *

from .time import calc_time_from_date

from typing import List, Dict, Set

class SimpleProblem(BaseException):
    def __init__(self, s):
        self.s = s


def update_items_from_json(data, g, et, fields,
                *,
                func_prefilter=lambda x: True,
                func_postfilter=lambda x: True,
                func_before_insert=lambda x: x,
                func_after_insert=lambda z,item, new_item: None):

    opts = [(key,old_key) for key,(old_key,new_type,purpose) in fields.items() if purpose == "key"]
    assert len(opts) == 1
    key_field, old_key_field = opts[0]

    problems = []
    changes = []

    with Transaction(g):
        entities = g | instances[now][et]
        for item_ind,orig_item in enumerate(data):
            try:
                if old_key_field not in orig_item:
                    raise SimpleProblem(f"Main key '{old_key_field}' not found in entry {item_ind}. Skipping")

                if not func_prefilter(orig_item):
                    continue
                item = {key: update_items_convert_value(g, orig_item, details)
                        for key,details in fields.items()}
                if not func_postfilter(item):
                    continue
                item = func_before_insert(item)

                # Find item
                if key_field not in item:
                    raise SimpleProblem(f"Main key '{key_field}' not found in entry {item_ind} after transformation. Skipping")

                entity_key = item[key_field]
                opts = entities | filter[lambda x: x >> RT(key_field) | value == entity_key]
                if len(opts) == 0:
                    entity = instantiate(et, g)
                    entities += ZefRefs([entity])
                    new_item = True
                elif len(opts) > 1:
                    raise SimpleProblem(f"Entry {item_ind}: unexpected multiple {et} with the same key {old_key_field} of {entity_key} found. Database problem detected.")
                else:
                    entity = opts | only
                    new_item = False
                    modified_field = False

                # Add fields in
                complained = False
                for key,details in fields.items():
                    old_key,new_type,purpose = details
                    try:
                        if purpose in ["static field", "dynamic field", "key"]:
                            val = item[key]
                            if new_item:
                                entity | fill_or_attach[RT(key), val]
                            else:
                                ae = entity >> O[RT(key)]
                                if ae is None or val != ae | value:
                                    modified_field = True
                                    if purpose in ["static field","key"] and not complained:
                                        problems.append(f'Static field "{key}" which should not change, has been updated to new value for item {et} with key {entity_key} - continuing with update anyway')
                                        complained = True
                                    entity | fill_or_attach[RT(key), val]
                        elif purpose == "calc":
                            # do nothing, this was just for postprocessing or filter or ...
                            pass
                        else:
                            raise Exception(f"Unknown purpose {purpose}")
                    except Exception as exc:
                        raise SimpleProblem(f"Unknown error occurred for key {key} in entry {item_ind}.")
                    

                ret = func_after_insert(entity, item, new_item)
                if ret is True:
                    modified_field = True

                if new_item:
                    changes.append(f"Added new {et} with key {entity_key}")
                elif modified_field:
                    changes.append(f"Modified existing {et} with key {entity_key}")
                else:
                    changes.append(f"Identical existing {et} with key {entity_key}")
                    

            except KeyboardInterrupt:
                raise
            except SimpleProblem as exc:
                problems.append(exc.s)
            except BaseException as exc:
                import traceback
                print(exc)
                print(traceback.format_exc())

                # Don't include the exception in the REST return
                # problems.append(f"Unknown exception: original key {orig_item[old_key_field]}, exc={exc}")
                problems.append(f"Unknown exception for entry {item_ind}")
                problems.append(f"Exception was {exc} with traceback {traceback.format_exc()}")

        return entities,problems,changes
    

        
def update_items_convert_value(g, item, details):
    old_key,new_type,purpose = details

    if old_key is None:
        # This must be handled in some post processing
        return None

    val = item[old_key]

    if new_type == str:
        if val is None:
            return ""
        else:
            return str(val)
    elif new_type == float:
        return float(val)
    elif new_type == int:
        return int(val)
    elif type(new_type) == str and new_type.startswith("EN."):
        en_type = new_type[len("ET."):]
        return EN(en_type, val)
    elif new_type == "date":
        if val is None:
            # If val is None then this will trigger an error later if it isn't handled in some post processing..
            return None
        return calc_time_from_date(g, val)
    elif new_type == "copy":
        return val
    else:
        raise Exception("Unknown new_type {new_type}")





######################################################
# * ZefSpec helper functions
#----------------------------------------------------

def AnyT(relent):
    if BT(relent) == BT.ENTITY_NODE:
        return ET(relent)
    elif BT(relent) == BT.ATOMIC_ENTITY_NODE:
        return AET(relent)
    elif BT(relent) == BT.RELATION_EDGE:
        return RT(relent)
    else:
        return BT(relent)


def zef_spec_check_relent(types, relent, errors, seen_relents):
    if AnyT(relent) not in types.keys():
        return
    if any(z == relent | to_ezefref for z in seen_relents):
        return
    seen_relents.append(relent | to_ezefref)

    relent_details = types[AnyT(relent)]
    zef_spec_check_relent_details(types, relent, relent_details, errors, seen_relents)


def zef_spec_check_relent_details(types, relent, relent_details, errors, seen_relents):
    relent_summary = f"relent {relent|uid} of type {AnyT(relent)}"
    to_check = list(relent | outs)
    additional = "no"

    for (key,rel_requirements) in relent_details.items():
        if type(key) == str:
            if key == "additional":
                assert additional in ["no", "allowed"]
                additional = rel_requirements
            elif key == "target":
                zef_spec_check_relation_details(types, relent, rel_requirements, errors, seen_relents)
                zef_spec_check_relent(types, relent | target, errors, seen_relents)
        elif type(key) == RelationType or type(key) == type(L) or type(key) == type(O):
            if type(key) == RelationType:
                rt = key
            elif type(key) == type(L):
                rt = key.data
            elif type(key) == type(O):
                rt = key.data
            opts = []
            for rel in list(to_check):
                if RT(rel) == rt:
                    opts.append(rel)
                    to_check.remove(rel)
            if type(key) == RelationType:
                if len(opts) > 1:
                    errors.append(f"Too many {key} for {relent_summary}")
                    continue
                elif len(opts) == 0:
                    errors.append(f"No {key} for {relent_summary}")
                    continue
            elif type(key) == type(O):
                if len(opts) > 1:
                    errors.append(f"Too many {key} for {relent_summary}")
                    continue

            for relation in opts:
                zef_spec_check_relation_details(types, relation, rel_requirements, errors, seen_relents)
                zef_spec_check_relent(types, relation | target, errors, seen_relents)

    if additional == "no":
        if len(to_check) > 0:
            extra_rts = set(RT(rel) for rel in to_check)
            errors.append(f"Extra unspecified relations on {relent_summary}: {extra_rts}")

    
    

            
def eq(a,b):
    if type(a) != type(b):
        return False
    return a == b


def noteq(a,b):
    return not eq(a,b)


def zef_spec_check_relation_details(types, relation, details, errors, seen_relents):
    import types as TYPES
    if type(details) not in [tuple,list]:
        details = [details]
    relation_summary = f"relation {relation|uid} of type {RT(relation)}"
    for requirement in details:
        if type(requirement) == TYPES.FunctionType:
            if not requirement(relation):
                errors.append(f"Function predicate failed for {relation_summary}")
        elif type(requirement) == dict:
            zef_spec_check_relent_details(types, relation, requirement, errors, seen_relents)
        else:
            if noteq(AnyT(relation | target), requirement):
                errors.append(f"Target should be ({requirement}) for {relation_summary} but was {AnyT(relation | target)}")
            

def zef_spec_check_structure(types, start_points, frame=None):
    errors = []
    seen_relents = []
    
    to_do = list(start_points)
    while to_do:
        item = to_do.pop()
        if type(item) == EntityType or type(item) == RelationType:
            if frame is None:
                raise TypeError("Need to pass a frame when giving ET/RT types.")
            to_do.extend(Graph(frame) | instances[frame][item])
        else:
            zef_spec_check_relent(types, item, errors, seen_relents)
    return len(errors) == 0, '\n'.join(errors)


################################
# * Merge helpers
#------------------------------
"""
    all_connected_entities(uzrs)    
Find all entities that would complete the set from `uzrs` such that no relation
has a source/target outside of the new set.
"""
def all_connected_entities(zrs):
    # FIXME: Working in a list for now, need to change to EZefRefs later
    all_relents = set(zrs)
    todo = list(zrs | filter[is_a[RT]] | collect)
    while todo:
        item = todo.pop()
        all_relents.add(item)
        if is_a(item, RT):
            todo += [source(item), target(item)]

    return ZefRefs(list(all_relents))


def add_connecting_relations(zs):
    todo = list(zs)
    full_set = set(zs)
    while todo:
        this = todo.pop()
        rels = this | ins_and_outs | filter[lambda z: z | source in full_set and z | target in full_set]
        for rel in rels:
            if rel not in full_set:
                full_set.add(rel)
                todo.append(rel)
                
    return ZefRefs(list(full_set))

def add_connected_entities(zs):
    full_set = set(zs)
    todo = list(ZefRefs(zs) | filter[lambda z: z | BT == BT.RELATION_EDGE])
    while todo:
        this = todo.pop()
        full_set.add(this)
        if this | BT == BT.RELATION_EDGE:
            todo.append(this | source)
            todo.append(this | target)
                
    return ZefRefs(list(full_set))
##################################
# * Import helpers
#--------------------------------


# More documentation needed here. Simple comments follow for the funcs:
#
# func_prefilter(raw) - passed the raw json entry for early filtering.
# func_postfilter(item) - passed the dictionary after automatic processing
# func_before_insert(item) - apply any modifications to the dictionary for manual processing (should return a dictionary)
# func_after_insert(zefref, item, new_item) - passed the entity, dictionary and bool "new_item" after this entry has been processed



##############################
# * Old things
#----------------------------

def cprint(x):
    os.write(1, str.encode(str(x) + '\n'))



##############################
# * Inject
#----------------------------
def instantiate_ae_from_val(val, g)->ZefRef:
    ae, = [val] | g | run
    return ae


def inject(d: any, g)->ZefRef:    
    """Function to import / inject existing data formats into a zefDB graph. 
    This may be extended to deal with data types beyond json (?).
    use as:
    d = {
        'zef_type': 'Cat',
        'zef_tag': 'small cat',
        'Name': 'Naomi',
        'Mother':{
            'zef_type': 'Cat',
            'zef_tag': 'big cat',
            'Name': 'Luna',
            'Age': 16,
            'favorite_foods': ['Tuna', 'Salmon', 42]
        },
        'Age': 15,
        'Weight': 4.5,
    }
    g = Graph()
    z = inject(d, g)    


    Note that we cannot express graph structures with multiple edges of the same type
    coming out of a single node as a dictionary: the keys have to be unique.
    """
    with Transaction(g):
        if type(d) in [str, float, bool, int]:
            return instantiate_ae_from_val(d, g)        
        
        if isinstance(d, list):
            my_list = instantiate(ET.List, g)
            prev_edge = None
            for el in d:                
                ed = instantiate(
                        my_list,
                        RT.ListElement,
                        inject(el, g),
                        g
                    )
                if prev_edge is not None:
                    instantiate(prev_edge, RT.Next, ed, g)
                prev_edge = ed                
            return my_list

        if isinstance(d, dict):
            et = ET(d.get('zef_type', '_unspecified'))
            z = instantiate(et, g)
            if 'zef_tag' in d:
                tag(z, d['zef_tag'])
            for rt, val in d.items():
                if rt not in ('zef_type', 'zef_tag'):
                    instantiate(
                        z,
                        RT(rt),
                        inject(val, g),
                        g
                    )            
            return z

######################################
# * Collect subgraph
#------------------------------------

def collect_subgraph(z_start, rels: List[Dict]):
    """
    z_start: elements in the initial subgraph, can be either ZefRefs or EZefRefs
    rels: [
            {'step': (ET.Schedule, RT.ScheduleCleaningTypeConnection, ET.CleaningType), 'step_direction': 'from_source'},
            {'step': (ET.ProcessOrder, RT.ProcessOrderShardableOperationInstanceConnection, ET.ShardableOperationInstance), 'step_direction': 'from_target', 'rel_predicate': lambda z: False},
          ]

    The first tuple means: If I am considering an entity of type ET.Schedule,
    walk outwards (step_direction) along RT.ScheduleCleaningTypeConnection and include the entity
    attached as a target in the set if it is of type ET.CleaningType.

    If 'step_direction'=='from_target' then we start walking from the target RAE (third element).
    If a 'rel_predicate' key is included, the associated function is applied to the relation being traversed.

    Starting from z_start, iteratively expand the set (subgraph)
    by adding RAE on the frontier connected by any of the relations
    specified in rels.
    """
    
    def to_zefrefs_or_uzefrefs(zs):
        if len(zs) == 0:
            raise RuntimeError(
                'set passed into extend_frontier is empty. Cannot determine whether it is a ZefRef or EZefRef')
        return ZefRefs(list(zs)) if isinstance(list(zs)[0], ZefRef) else EZefRefs(list(zs))

    def extend_frontier(frontier: Set, rels: List[Dict]) -> Set:
        """return both the source/target attached, but also the relations if a triple is to be included """

        out_rels = ([(to_zefrefs_or_uzefrefs(frontier) | filter[d['step'][0]] > L[(d['step'][1])]) | flatten | filter[
            lambda z: rae_type(z | target) == d['step'][2]] | filter[d.get('rel_predicate', lambda z: True)] for d in
                     rels if d['step_direction'] == 'from_source'] | ops.flatten())
        in_rels = ([(to_zefrefs_or_uzefrefs(frontier) | filter[d['step'][2]] < L[(d['step'][1])]) | flatten | filter[
            lambda z: rae_type(z | source) == d['step'][0]] | filter[d.get('rel_predicate', lambda z: True)] for d in
                    rels if d['step_direction'] == 'from_target'] | ops.flatten())
        return set(out_rels + in_rels + [z | target for z in out_rels] + [z | source for z in in_rels])

    current_set = set(z_start)
    frontier = set(z_start)
    # this is gonna be mutaty style for now :(
    while frontier:
        candidates = set(extend_frontier(frontier, rels))
        frontier = candidates - current_set  # only keep the ones we have never seen
        current_set.update(frontier)  # add all new elements by mutating the existing set

    return to_zefrefs_or_uzefrefs(current_set)



##############################
# * timeit
#----------------------------
def timeit(f):
    """use as decorator on any fct for a very simple print output measuring the execution duration"""
    def wrapped(*args, **kwargs):
        import time
        t0=time.time()
        res = f(*args, **kwargs)
        t1=time.time()
        print(f"zefdb.timeit - duration for single run: {t1-t0} seconds")
        return res
    return wrapped












# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from zef import *
from zef.ops import *
from .ariadne_utils import gqlify, de_gqlify


def parse_list_et(list_et):
    aet_values = []
    arguments = list_et >> L[RT.argument] | collect
    at = ""
    for arg in arguments:
        if BT(arg) == BT.ENTITY_NODE:
            return "", de_gqlify(value(arg >> RT.Name), True)
        else:
            aet_type = str(AET(arg | to_ezefref | collect))
            at = aet_type
            if at == "Bool":
                at = "Boolean"
            if value(arg):
                aet_values.append(value(arg))
            else:
                return "", at

    aet_values = "[" + ','.join([str(x) for x in aet_values]) + "]"
    return aet_values, at


def check_nonnullable(param):
    nonnullable = ""
    if length(param > L[RT(gqlify("nonNullable"))]) > 0:
        nonnullable = "!"
    return nonnullable


def check_implements(param):
    implements = ""
    if length(param > L[RT(gqlify("implements"))]) > 0:
        interface_name = de_gqlify(value(param >> RT(gqlify("implements")) >> RT.Name), True)
        implements = f"implements {interface_name}"
    return implements


def parse_queryparams_et(params_et):
    returned_list = []
    if length(params_et >> L[RT(gqlify("queryParams"))]) > 0:
        for param in params_et > L[RT(gqlify("queryParams"))] | collect:
            param_name = de_gqlify(value(param >> RT.Name), False)
            nonnullable = check_nonnullable(param)
            if BT(param | target | collect) == BT.ATOMIC_ENTITY_NODE:
                aet_type = str(AET(param | target | collect))
                aet_value = param | target | value | collect
                if aet_type == "Bool":
                    aet_type = "Boolean"
                if aet_value:
                    returned_list.append(f"{param_name}: {aet_type}{nonnullable} = {aet_value}")
                else:
                    returned_list.append(f"{param_name}: {aet_type}{nonnullable}")
            else:
                if de_gqlify(str(ET(param | target | collect)), False) == "list":
                    aet_values, et_type = parse_list_et(param | target | collect)
                    if aet_values:
                        aet_values = " = " + aet_values
                    returned_list.append(f"{param_name}: [{et_type}]{nonnullable}{aet_values}")
                else:
                    et_type = de_gqlify(value(((param | target) >> RT.Name)), True)
                    returned_list.append(f"{param_name}: {et_type}{nonnullable}")

        return returned_list
    else:
        return None


def generate_fields(et, schema_out):
    for rt in (et > L[RT(gqlify("field"))] | collect):
        nonnullable = check_nonnullable(rt)
        field_name = de_gqlify(value(rt >> RT.Name), False)
        field_return = rt | target | collect
        blob_type = BT(field_return | to_ezefref | collect)
        query_params = parse_queryparams_et(rt)

        def inject_params(qp):
            if qp:
                return f"({','.join(qp)})"
            else:
                return ""

        if blob_type == BT.ATOMIC_ENTITY_NODE:
            aet_type = str(AET(field_return | to_ezefref | collect))
            aet_value = field_return | value | collect
            if aet_type == "Bool":
                aet_type = "Boolean"
            if aet_value:
                aet_value = f" = {aet_value}"
            else:
                aet_value = ""

            schema_out += f'  {field_name}{inject_params(query_params)}: {aet_type}{nonnullable}{aet_value}\n'
        else:
            if de_gqlify(str(ET(field_return)), False) == "list":
                aet_values, et_type = parse_list_et(field_return)
                if aet_values:
                    aet_values = " = " + aet_values
                schema_out += f'  {field_name}{inject_params(query_params)}: [{et_type}]{nonnullable}{aet_values}\n'
            else:
                schema_out += f'  {field_name}{inject_params(query_params)}: {de_gqlify(value(field_return >> RT.Name),True)}{nonnullable}\n'
    schema_out += "}\n"
    return schema_out


def generate_enum_values(en, schema_out):
    for field in (en >> L[RT(gqlify("field"))] | collect):
        en_value = (field | value).enum_value
        schema_out += f'  {en_value}\n'
    schema_out += "}\n"
    return schema_out


def generate_api_schema_string(schema_root):
    types = schema_root >> L[RT(gqlify("type"))] | collect
    inputs = schema_root >> L[RT(gqlify("input"))] | collect
    interfaces = schema_root >> L[RT(gqlify("interface"))] | collect
    scalars = schema_root >> L[RT(gqlify("scalar"))] | collect
    enums = schema_root >> L[RT(gqlify("enum"))] | collect

    schema_out = ""
    for et in types:
        implements = check_implements(et)
        object_type = de_gqlify(value(et >> RT.Name), True)
        schema_out += f'type {object_type} {implements} {"{"}\n'
        schema_out = generate_fields(et, schema_out)

    for et in inputs:
        object_type = de_gqlify(value(et >> RT.Name), True)
        schema_out += f'input {object_type} {"{"}\n'
        schema_out = generate_fields(et, schema_out)

    for et in interfaces:
        object_type = de_gqlify(value(et >> RT.Name), True)
        schema_out += f'interface {object_type} {"{"}\n'
        schema_out = generate_fields(et, schema_out)

    for et in scalars:
        object_type = de_gqlify(value(et >> RT.Name), True)
        schema_out += f'scalar {object_type}\n'

    for et in enums:
        object_type = de_gqlify(value(et >> RT.Name), True)
        schema_out += f'enum {object_type} {"{"}\n'
        schema_out = generate_enum_values(et, schema_out)

    return schema_out


def generate_api_schema_file(schema_root, destination):
    f = open(destination, "w+")
    f.write('generated_schema = """\n')
    schema_string = generate_api_schema_string(schema_root)
    f.write(schema_string)
    f.write('"""\n')
    f.close()



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...core import *
from ...pyzef.zefops import assign_value, now, L, first

def auto_type_from_val(val):
    if type(val) == str:
        return AET.String
    elif type(val) == int:
        return AET.Int
    elif type(val) == bool:
        return AET.Bool
    elif type(val) == float:
        return AET.Float
    elif type(val) == ZefEnumValue:
        return getattr(AET.Enum, val.enum_type)
    elif type(val) == QuantityInt:
        return getattr(AET.QuantityInt, val.unit.enum_value)
    elif type(val) == QuantityFloat:
        return getattr(AET.QuantityFloat, val.unit.enum_value)
    elif type(val) == Time:
        return AET.Time
    elif type(val) == ZefRef or type(val) == EZefRef:
        return ZefRef
    else:
        raise TypeError(f"Unknown type for property: {type(val)}, {val}")

class Attach():    
    def __init__(self, args=()):
        # future API:
        #     my_machines | attach[RT.Age, [6,1,9,4]]                    # one ZefRef, multiple fields
        #     my_machine | attach[ [(RT.Age, 54), (RT.Color, 'red')] ]   # multiple args per ZefRef       

        # Note: fields is stored, after parsing, as a list of 3 items, rt,aet,val
        # If val is not given, then it is set to None.
        # As a special case, an aet of ZefRef (the type not an instance) means create a link only.

        if args==():
            self.fields = []
            return

        if type(args) == list:
            fields = args
        elif len(args) >= 2 and len(args) <= 3 and isinstance(args[0], RelationType):
            fields = [args]
            pass
        else:
            raise ValueError("Incorrect use of attach. Should be either an RT, or an RT/val pair, or a list.")

        parsed = []
        for field in fields:
            assert len(field) >= 2 and len(field) <= 3 and isinstance(field[0], RelationType)
            rt = field[0]

            def is_compatible_type(val, ae_type)->bool:            
                return auto_type_from_val(val) == ae_type
        
            # ----- two additional args -----
            if len(field) == 3:
                aet,val = field[1:]

                assert isinstance(aet, AtomicEntityType)
                if not auto_type_from_val(val) == aet:
                    raise TypeError(f'in attach: the type of the specified value "{args[1]}" does not match the specified type "{args[0]}"')
                parsed.append( (rt, aet, val) )
            else:
                # ----- one additional args -----
                item = field[1]
                if isinstance(item, AtomicEntityType):
                    parsed.append( (rt, item, None) )
                else:
                    parsed.append( (rt, auto_type_from_val(item), item) )
        self.fields = parsed

    def __getitem__(self, args):  
        # if we pass multiple args via [], they are always grouped into a single tuple assigned to the first arg after self        
        return type(self)(args)
    
    def __call__(self, z):
        g = Graph(z)
        with Transaction(g):
            for field in self.fields:
                rt,aet,val = field
                # Need "is" here to avoid failing to find a == method.
                if aet is ZefRef:
                    assert val is not None
                    instantiate(z, rt, val, g)            
                else:
                    my_ae = instantiate(aet, g)
                    if val is not None:
                        assign_value(my_ae, val)
                    instantiate(z, rt, my_ae, g)            
        return z | now   # TODO: confirm that projecting to the latest time slice makes sense here
    
    def __ror__(self, z):
        return self(z)

attach = Attach()


class FillOrAttach(Attach):
    """
    Check whether a field exists.
        a) if it does, assign the value to the target AET
        b) if it does not, instantiate the field and assign the value to the target AET
        
    smaple:
        z_person | fill_or_attach[RT.Age, 42]
    """
    def __init__(self, args=()):
        super().__init__(args)

        aet_rts = []
        for field in self.fields:
            rt,aet,val = field
            # Need "is" here to avoid failing to find a == method.
            if aet is not ZefRef:
                aet_rts.append(rt)

        import collections
        counts = collections.Counter(aet_rts)
        if any(count > 1 for count in counts.values()):
            raise TypeError("Can't have multiple RTs for an AET in fill_or_attach")
        
    def __call__(self, z):
        z = now(z)
        g = Graph(z)
        with Transaction(g):
            for field in self.fields:
                rt,aet,val = field
                # Need "is" here to avoid failing to find a == method.
                if aet is ZefRef:
                    # Always append new ZefRefs, these can never be filled.
                    assert val is not None
                    instantiate(z, rt, val, g)            
                else:
                    opts = z >> L[rt]
                    if len(opts) == 0:
                        my_ae = instantiate(aet, g)
                        if val is not None:
                            assign_value(my_ae, val)
                        instantiate(z, rt, my_ae, g)            
                    else:
                        if val is not None:
                            if len(opts) > 1:
                                raise Exception(f"More than one RT found for {rt} in fill_or_attach, going to assign to the first RT found only.")
                            assign_value(opts | first, val)
        return z | now   # TODO: confirm that projecting to the latest time slice makes sense here
    
    def __ror__(self, z):
        return self(z)

fill_or_attach = FillOrAttach()



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .fx_types import _Effect_Class
from ...pyzef.internals import login, logout

def zefhub_login_handler(eff: _Effect_Class):
    login()
    return {}

def zefhub_logout_handler(eff: _Effect_Class):
    logout()
    print("Warning, after logging out all graphs will have been unloaded. Any accesses of a ZefRef referencing these old graphs will cause a segmentation fault!")
    return {}


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from zefdb import *
from zefdb.zefops import *

##############################
# * Utils
#----------------------------

def neighbours(l):
    for i in range(1,len(l)):
        yield l[i-1],l[i]


def same_entity(x,y):
    return x|to_ezefref == y|to_ezefref
##############################
# * Sets
#----------------------------

# This is nothing special, L[RT.ListElement] is nearly identical. Just convenience functions making assumptions.

def set_toggle(s, item, rt=RT.ListElement):
    opts = (s|now) >> L[rt] | filter[lambda x: same_entity(x,item)]
    if len(opts) == 0:
        instantiate(s, rt, item, Graph(s))
    else:
        opts | only | terminate

def set_add(s: ZefRef, item: ZefRef, rt=RT.ListElement):
    opts = (s|now) >> L[rt] | filter[lambda x: same_entity(x,item)]
    if len(opts) == 0:
        instantiate(s, rt, item, Graph(s))

def set_remove(s: ZefRef, item: ZefRef, rt=RT.ListElement):
    opts = ((s|now) > L[rt]) | filter[lambda x: same_entity(x|target,item)]
    for opt in opts:
        terminate(opt)

def in_set(s, item, rt=RT.ListElement):
    opts = (s|now) >> L[rt] | filter[lambda x: same_entity(x,item)]
    return len(opts) >= 1

def set_empty(s, rt=RT.ListElement):
    for rel in (s|now) > L[rt]:
        terminate(rel)



# z1 | attach[RT.a, z2]                           # returns z1
# z_list | append_to_zef_list[z_my_element]       # returns the z_list
# z_set | add_to_zef_set[z_my_element]            # returns the z_set
# z_set | add_to_zef_set([z_my_element1, z_my_element)]            # returns the z_set
# z_list | contains[z1]


##############################
# * List
#----------------------------

# Lists have order, designated by the RT.Next relations between the relations RTListElement.

def list_start(l, rt=RT.ListElement, return_ind=False):
    opts = l > L[rt]
    if len(opts) == 0:
        if return_ind:
            return None,None
        else:
            return None
    
    start_rel = opts | first
    ind = 0
    while True:
        opts = start_rel << L[RT.Next]
        if len(opts) == 0:
            break
        if start_rel == opts | only:
            raise Exception("Invalid list state")
        start_rel = opts | only
        ind += 1

    # start = start_rel | target
    if return_ind:
        return start_rel,ind
    else:
        return start_rel

def list_end(l, rt=RT.ListElement):
    opts = l > L[rt]
    if len(opts) == 0:
        return None
    
    end_rel = opts | first
    while True:
        opts = end_rel >> L[RT.Next]
        if len(opts) == 0:
            break
        if end_rel == opts | only:
            raise Exception("Invalid list state")
        end_rel = opts | only

    # return end_rel | target
    return end_rel
        
def list_nth(l, ind, rt=RT.ListElement):
    assert ind >= 0

    start_rel,first_ind = list_start(l, rt, return_ind=True)
    if first_ind <= ind:
        search_ind = first_ind
        search_rel = (l > L[rt]) | first
    else:
        search_ind = 0
        search_rel = start_rel

    while search_ind < ind:
        search_rel = search_rel >> RT.Next
        search_ind += 1

    return search_rel

def list_as_ZefRefs(l, rt=RT.ListElement, invalid_fallback=False):
    start_rel = list_start(l, rt)
    if start_rel is None:
        return ZefRefs([], l|tx)

    cur_rel = start_rel

    out = []
    while True:
        out.append(cur_rel)
        opts = cur_rel >> L[RT.Next]
        if len(opts) == 0:
            break
        cur_rel = opts | only

    if len(out) != len(l > L[rt]):
        if invalid_fallback:
            if len(out) >= 2:
                # Note: This warning doesn't get triggered, if the "start_rel"
                # is isolated, even if other items have RT.Next on them.
                import logging
                logging.warning(f"List {l} has some RT.Next but not enough to complete the list")
            return l > L[rt]
        raise Exception(f"Warning: not all {rt} attached to {l} have {RT.Next} connecting them to make a valid list.")
    return ZefRefs(out)

def list_empty(l, rt=RT.ListElement):
    with Transaction(Graph(l)):
        for z in (l | now) > L[rt]:
            terminate(z)

def list_el_fill(l, items, rt=RT.ListElement, g=None):
    # Note: the "now" is very important, otherwise we can be adding links based on an old picture
    g = Graph(l) if g is None else g
    with Transaction(g):
        l = l | now
        list_empty(l)

        rels = []
        for item in items:
            rels.append(instantiate(l, rt, item, g))
        # can we write rels = [instantiate(l, rt, item, g) for item in items]  ?

        for prev,after in neighbours(rels):
            instantiate(prev, RT.Next, after, g)

def list_el_pushfront(l, item, rt=RT.ListElement, g=None):
    # Note: the "now" is very important, otherwise we can be adding links based on an old picture
    l = l | now
    g = Graph(l) if g is None else g
    with Transaction(g):
        start_rel = list_start(l, rt)
        rel = instantiate(l, rt, item, g)
        if start_rel is not None:
            instantiate(rel, RT.Next, start_rel, g)

def list_el_pushback(l, item, rt=RT.ListElement, g=None):
    # Note: the "now" is very important, otherwise we can be adding links based on an old picture
    l = l | now
    g = Graph(l) if g is None else g
    with Transaction(g):
        end_rel = list_end(l, rt)
        rel = instantiate(l, rt, item, g)
        if end_rel is not None:
            instantiate(end_rel, RT.Next, rel, g)

def list_pop(l, rel, g=None):
    g = Graph(l) if g is None else g
    with Transaction(g):
        l = l|now
        rel = rel|now
        following = rel >> O[RT.Next]
        preceeding = rel << O[RT.Next]
        terminate(rel)
        if following is not None and preceeding is not None:
            I(preceeding, RT.Next, following)

    

##############################
# * Matrix
#----------------------------

# Matrices have a list of rows, list of columns, and values connecting these.

def create_matrix(rows, cols, g, add_labels=True):
    I = lambda *args: instantiate(args, g)

    with Transaction(g):
        matrix = I(ET.Matrix)

        row_rels = []
        for row in rows:
            z = I(ET.Row)
            if add_labels:
                I(z, RT.Label, row)
            row_rels.append(I(matrix, RT.Row, z))

        for prev,after in neighbours(row_rels):
            I(prev, RT.Next, after)

        col_rels = []
        for col in cols:
            z = I(ET.Column)
            col_rels.append(I(matrix, RT.Column, z))
            if add_labels:
                I(z, RT.Label, col)
            
        for prev,after in neighbours(col_rels):
            I(prev, RT.Next, after)

    return matrix

def matrix_rows_as_list(matrix):
    return list_as_ZefRefs(matrix, RT.Row)
def matrix_cols_as_list(matrix):
    return list_as_ZefRefs(matrix, RT.Column)
    
def matrix_row(matrix, ind):
    return list_nth(matrix, ind, RT.Row)
def matrix_col(matrix, ind):
    return list_nth(matrix, ind, RT.Column)

def matrix_val(matrix, i, j):
    row = matrix_row(matrix, i)
    col = matrix_col(matrix, j)
    return connection(row, RT.Value, col)

def matrix_vals_as_lists(matrix):
    rows = list_as_ZefRefs(matrix, RT.Row)
    cols = list_as_ZefRefs(matrix, RT.Column)

    total = []
    for row in rows:
        this_list = []
        for col in cols:
            item = connection(row, RT.ListElement, col) >> RT.Value
            this_list.append(item)
        total.append(this_list)

    return total

def fill_matrix(matrix, items, g=None):
    # Note: the "now" is very important, otherwise we can be adding links based on an old picture
    matrix = matrix | now
    rows = matrix_rows_as_list(matrix)
    cols = matrix_cols_as_list(matrix)

    g = Graph(matrix) if g is None else g
    with Transaction(g):
        for (i,row) in enumerate(rows):
            row > L[RT.Value] | terminate
            for (j,col) in enumerate(cols):
                # connections(row, RT.Value, col) | terminate
                I(row, RT.ListElement, col) | make_fields_(RT.Value, items[i][j])




# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest  # pytest takes ages to run anything as soon as anything from zef is imported
from zef import *
from zef.ops import *
import zef

class MyTestCase(unittest.TestCase):
    def test_networkx(self):
        from networkx import DiGraph
        import networkx as nx
        # from networkx.algorithms.components import strongly_connected_components

        from zef.experimental.networkx import ProxyGraph

        g = Graph()
        zs = [ET.Machine]*42 | g | run
        for i,j in [(0,1), (1,2), (2,5), (5,10), (10,0),
                    (5,6), (6,7), (7,8), (8,6)]:
            (zs[i], RT.UsedBy, zs[j]) | g | run

        r = [
            (ET.Person["alex"], [(RT.FirstName, "Alex"),
                                (RT.Status, EN.Status.Developer),
                                (RT.ZefAge, 1)]),
            (ET.Person["bob"], [(RT.FirstName, "Bob"),
                                (RT.Status, EN.Status.Zefer),
                                (RT.ZefAge, 0)]),
            (ET.Person["zach"], [(RT.FirstName, "Zach"),
                                (RT.Status, EN.Status.Nobody)]),

            (Z["alex"], RT.Knows["d-e"], Z["bob"]),
            (Z["d-e"], RT.From, "University"),

            (Z["alex"], RT.Knows["d-z"], Z["zach"]),
            (Z["d-z"], RT.From, "Nowhere"),
        ] | transact[g] | run


        dg = ProxyGraph(now(g), ET.Person, RT.Knows)
        self.assertEqual(set(dg.nodes) | map[lambda x: x.z] | func[set] | collect,
                         set(g | now | all[ET.Person] | collect))
        self.assertEqual(set(dg.edges) | map[map[lambda x: x.z] | func[tuple]] | func[set] | collect,
                         set([(r["alex"], r["bob"]),
                              (r["alex"], r["zach"])]))
        self.assertEqual(dg[r["alex"]][r["bob"]]["From"], "University")
        self.assertEqual(dg.nodes[r["alex"]]["ZefAge"], 1)
        self.assertEqual(dg.edges[r["d-e"]]["From"], "University")

        dg2 = ProxyGraph(now(g), ET.Machine, RT.UsedBy)
        list(nx.strongly_connected_components(dg2))

        ug = ProxyGraph(now(g), ET.Person, RT.Knows, undirected=True)
        list(nx.connected_components(ug))

        nx.all_pairs_node_connectivity(ug)
        nx.all_pairs_node_connectivity(dg)

        nx.node_connectivity(ug)
        nx.node_connectivity(dg)

        nx.node_connectivity(ug, list(dg.nodes)[1], list(dg.nodes)[2])
        nx.node_connectivity(dg, list(dg.nodes)[1], list(dg.nodes)[2])

        last_two = [r["bob"], r["zach"]]
        dg_sg = dg.subgraph(last_two)
        ug_sg = ug.subgraph(last_two)
        self.assertEqual(nx.node_connectivity(ug), 1)
        self.assertEqual(nx.node_connectivity(ug_sg), 0)

        nx.k_components(ug)
        nx.k_components(dg2.to_undirected())

        # Can't work - tries to construct a graph itself
        # nx.maximum_branching(dg)
        nx.maximum_branching(dg.to_native())

        # Doesn't work - tries to use _pred
        # nx.average_clustering(dg)
        nx.average_clustering(dg.to_native())

        nx.diameter(ug)

        (r["bob"], RT.Knows, r["zach"]) | g | run
        ug_complete = ProxyGraph(now(g), ET.Person, RT.Knows, undirected=True)

        # Doesn't work - tries to construct a graph itself
        # nx.minimum_spanning_tree(ug_complete)
        nx.minimum_spanning_tree(ug.to_native())

        nx.degree_centrality(dg)
        nx.degree_centrality(ug)

        list(nx.chain_decomposition(ug_complete))

        nx.is_chordal(dg2.to_undirected())

        list(nx.enumerate_all_cliques(ug))

        nx.greedy_color(dg)
        nx.greedy_color(dg2)
        nx.greedy_color(ug)

        for k in range(2,4):
            list(nx.algorithms.community.k_clique_communities(dg2.to_undirected(), k))

        nx.shortest_path(ug)
        nx.shortest_path(dg)
        nx.shortest_path(dg2)

if __name__ == '__main__':
    unittest.main()


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    "merge",
]

from ..pyzef.main import Graph

# Override merge so that we may pass in a dict into the C++ code for GraphDeltas
def merge(obj, g : Graph, fire_and_forget : bool = False):
    from ..pyzef.main import merge as orig_merge
    from .serialization import serialize, deserialize

    serialized = serialize(tuple(obj))
    rs =  orig_merge(serialized, g, fire_and_forget)
    receipt = deserialize(rs)
    return receipt




# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .. import *
from .._ops import *
from .. import VT
import random

##############################################################
# * ZefOp chain yo output implementations
def pad_side(s: str, padding: int, left: bool, desired_length: int) -> str:
    n = desired_length - len(s)
    if n<0:
        raise RuntimeError('pad_side is only well defined if the desired length is larger than the length of the string to pad')
    if left:
        return f"{' ' * padding}{s}{' ' * (desired_length - len(s) - padding)}"
    else:
        return f"{' ' * (desired_length - len(s) - padding)}{s}{' ' * padding}"


def inject_in_line(line: list, title: str):
    if not title: return line
    if len(title) > len(line) - 6: title = title[:len(line) - 10] + "..."
    title = f" {title} "
    return [*line[:3] , title, *line[3+len(title):]] 


@func
def make_box(content, label = ""):
    max_width = content | map[length] | max | collect
    width  = max_width + 30
    padding_v = 2
    padding_h = 8

    up     = ("─" | repeat |  take[width] | prepend["┌"] | append["┐"] | collect)  
    up     = inject_in_line(up, label) | join | collect

    middle = " " | repeat |  take[width] | prepend["│"] | append["│"] | join | repeat | take[padding_v] | join["\n"] | collect
    
    middle_content = content | map[lambda l: f"|{pad_side(l, padding_h, True, width)}|"] | join["\n"] | collect
    
    bottom = "─"  | repeat |  take[width] | join | prepend["└"] | append["┘"] | join | collect
    return [up, middle, middle_content, middle, bottom]  | join["\n"] | collect

@func
def make_box_details(op_chain, out_types, op_purity):
    return f"""
!! This is an incomplete/incorrect output !!
Operator 🔗:    {op_chain}
Signature:      {out_types[0]} -> {out_types[-1]}
Purity:         {random.SystemRandom().choice(['🌿 pristenly pure', '👹 possibly mutating'])}\n\n"""


@func
def make_operator_repr(op, out_type, op_purity):
    op          = ZefOp((op, ))
    purity      = f"│ {op_purity}"
    operator    = f"│ {str(op)}"
    out_type    = f"{out_type}"
    return [purity, operator, "▼", out_type]


@func
def chain_output(op: ZefOp, out_types: list, op_purity: list) -> list:
    initial_type = first(out_types)
    content = (op.el_ops, out_types[1:], op_purity) | zip | for_each[lambda x: make_operator_repr(*x)] | concat | prepend[str(initial_type)] | collect
    return content


def make_op_chain_ascii_output(op_chain):
    out_types = [VT.Any for _ in range(len(op_chain.el_ops) + 1)]
    op_purity = [random.SystemRandom().choice('🌿👹') for _ in range(len(op_chain.el_ops))]


    return (
        LazyValue(op_chain) 
        | chain_output[out_types][op_purity]
        | make_box["Data pipeline"] 
        | prepend[make_box_details(op_chain, out_types, op_purity)]
        | collect
    )


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def c_merge_handler(g, serialized_delta):
    """Don't call this explicitly. Only for the zefdb core."""

    from ..serialization import serialize, deserialize
    from .._ops import run, transact

    # TODO: Double check primary role here.

    commands = deserialize(serialized_delta)
    from zef.core.graph_delta import perform_transaction_commands
    receipt = perform_transaction_commands(commands, g)

    return serialize(receipt)

def register_merge_handler():
    from ...pyzef import internals
    internals.register_merge_handler(c_merge_handler)



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .fx_types import Effect
from ..error import Error
import yaml 
import json 
import toml
import  pandas as pd
import io
from ..image import Image

#TODO Docstring!

def read_localfile_handler(eff: Effect):
    filename  = eff.d["filename"]

    f = open(filename, "r")
    f = f.read()
    return {"content": f}


def load_localfile_handler(eff: Effect):
    filename  = eff.d["filename"]
    settings  = eff.d.get("settings", {})
    format    = eff.d.get("format", None)

    if not format:
        if "." not in filename: return Error.ValueError("filename is missing an extension and a format wasn't provided!", filename)
        else: format = filename[filename.rindex(".") + 1:]
    elif "." not in filename: filename = filename + f".{format}"

    f = open(filename, "r")
    content = f.read()

    if format in {"svg", "png", "jpg", "jpeg"}:
        content = bytes(content, "UTF-8")
        content = Image(content, format)
    elif format in {"yaml", "yml"}:
        content = yaml.safe_load(content)
    elif format == "toml":
        content = toml.loads(content)
    elif format == "csv":
        content = pd.read_csv(io.StringIO(content), **settings)
    elif format == "json":
        content = json.loads(content)

    return {"content": content, "format": format}


def save_localfile_handler(eff: Effect):
    """
    Example Effect

    Effect({
        'type':     FX.LocalFile.Save,
        'filename': fname,
        'content':  content,
        'settings': settings,
    })
    """

    content   = eff.d["content"]
    filename  = eff.d["filename"]
    settings  = eff.d.get("settings", {})

    if isinstance(content, Image):
        format = f'.{content.format}'
        if format not in {".svg", ".png", ".jpg", ".jpeg"}: return Error.ValueError(f'Image format needs to be one of these types {{".svg", ".png", ".jpg", ".jpeg"}} got {format} instead.')        
        filename = filename if format == (filename[-4:] or filename[-5:])  else filename + format
        import zstd
        with open(filename, 'wb') as f: f.write(zstd.decompress(content.buffer))
        
    elif "." in filename:
        format = filename[filename.rindex(".") + 1:]
        if format in {"yaml", "yml"}:
            with open(filename, 'w') as f: f.write(yaml.safe_dump(content))
        elif format == "toml":
            with open(filename, 'w') as f: f.write(toml.dumps(content))
        elif format == "csv":
            with open(filename, 'w') as f: f.write(content.to_csv(**settings))
        elif format == "json":
            with open(filename, 'w') as f: f.write(json.dumps(content))
        else:
            return Error.NotImplementedError(f'Unsupported writing to file type {format}!')
    else:
        filename += ".txt"
        with open(filename, 'w') as f: f.write(content)

    return {"filename": filename}


def write_localfile_handler(eff: Effect):
    """
    Example Effect

    Effect({
        'type':     FX.LocalFile.Write,
        'filename': fname,
        'content':  content,
    })
    """

    content   = eff.d["content"]
    filename  = eff.d["filename"]

    with open(filename, "w") as file: file.write(content)
    return {"filename": filename}

def system_open_with_handler(eff: Effect):
    """
    This will open the file using the system's default handler application.

    Example
    =======

    Effect({
            'type': FX.LocalFile.SystemOpenWith,
            'filepath': fname,
    })
    """

    import subprocess, os, platform
    if platform.system() == 'Darwin':       # macOS
        subprocess.Popen(('open', eff.d['filepath']))
    elif platform.system() == 'Windows':    # Windows
        os.startfile(eff.d['filepath'])
    else:                                   # linux variants
        subprocess.Popen(('xdg-open', eff.d['filepath']))


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import imgui
import sys
import yaml
from caseconverter import pascalcase

from ... import *
from ...ops import *

purpose_opts = ["entity", "field", "field_on", "ignore", "source", "target", "id"]
purpose_keys = {}
purpose_common_keys = ["name", "purpose", "data_type"]
purpose_keys["id"] = purpose_common_keys + ["ET"]
purpose_keys["entity"] = purpose_common_keys + ["ET", "RT"]
purpose_keys["field"] = purpose_common_keys + ["RT"]
purpose_keys["field_on"] = purpose_common_keys + ["RT", "target"]
purpose_keys["ignore"] = purpose_common_keys + []
purpose_keys["source"] = purpose_common_keys + ["ET"]
purpose_keys["target"] = purpose_common_keys + ["ET"]
purpose_keys["id"] = purpose_common_keys + ["RT"]

# TODO - allow ignoring of tables.
kind_opts = ["entity", "relation", "ignore"]
kind_common_keys = ["tag", "data_source", "cols", "kind"]
kind_keys = {}
kind_keys["entity"] = kind_common_keys + ["ID_col", "ET"]
kind_keys["relation"] = kind_common_keys + ["ID_col", "RT"]

##############################
# * Theming
#----------------------------

entity_background = (4/255, 32/255, 194/255)
relation_background = (4/255, 125/255, 91/255)

entity_foreground = (120/255, 133/255, 254/255)
special_foreground = (174/255, 79/255, 214/255)

active_step_color = (0.8, 0.1, 0.1)
done_step_color = (0.0, 0.0, 0.0)
done_step_text_color = (0.2, 0.2, 0.2)

##################################
# * Data wranglers
#--------------------------------


from dataclasses import dataclass
@dataclass
class Empty:
    pass

def maybe_dict_as_obj(v):
    if isinstance(v, dict):
        return dict_as_obj(v)
    if isinstance(v, tuple):
        return tuple([dict_as_obj(x) for x in v])
    if isinstance(v, list):
        return [dict_as_obj(x) for x in v]
    return v

def dict_as_obj(d):
    obj = Empty()
    for key,val in d.items():
        val = maybe_dict_as_obj(val)
        setattr(obj, key, val)
    return obj

def obj_as_dict(obj):
    if isinstance(obj, Empty):
        d = {}
        for key in keys(obj):
            val = getattr(obj, key)
            val = obj_as_dict(val)
            d[key] = val
        return d
    if isinstance(obj, list):
        return [obj_as_dict(x) for x in obj]
    return obj

def get(obj, key, default):
    if hasattr(obj, key):
        return getattr(obj, key)
    return default

def keys(obj):
    return [x for x in dir(obj) if x[0] != '_']

def get_col(cols, name):
    for col in cols:
        if col.name == name:
            return col

    return None

def get_col_name_from_purpose(cols, purpose):
    col = [x.name for x in cols if x.purpose == purpose]
    if col:
        return col[0]
    else:
        return None


def change_col_to_purpose(item, new_purpose):
    if new_purpose == "entity":
        item.ET = get(item, "ET", get(item, "RT", item.name))
        item.RT = get(item, "RT", get(item, "ET", item.name))
    elif new_purpose == "field":
        item.RT = get(item, "RT", get(item, "ET", item.name))
    elif new_purpose == "field_on":
        item.RT = get(item, "RT", get(item, "ET", item.name))
        item.target = get(item, "target", get(item, "ET", ""))
    elif new_purpose == "ignore":
        pass
    elif new_purpose == "source":
        item.ET = get(item, "ET", get(item, "ET", item.name))
    elif new_purpose == "target":
        item.ET = get(item, "ET", get(item, "ET", item.name))
    elif new_purpose == "id":
        item.RT = get(item, "RT", get(item, "RT", "ID"))
    else:
        raise Exception(f"Don't understand purpose: '{purpose}'")
    item.purpose = new_purpose

    if "data_type" in purpose_keys[new_purpose]:
        item.data_type = get(item, "data_type", "String")

    for key in set(keys(item)).difference(purpose_keys[new_purpose]):
        delattr(item, key)

    return item


##############################
# * Init
#----------------------------


def init_state(filepath):
    with open(filepath) as file:
        decl = yaml.safe_load(file)

    imgui.get_style().window_rounding = 0
    S = Empty()
    S.decl = dict_as_obj(decl)
    S.save_filename = filepath
    S.cur_step = 0
    return S

##############################
# * Render
#----------------------------



def render(S):
    if imgui.begin_main_menu_bar():
        clicked_save, _ = imgui.menu_item(
            "Save", "", False, True
        )
        clicked_quit, _ = imgui.menu_item(
            "Quit without saving", "", False, True
        )

        if clicked_save:
            with open(S.save_filename, "w") as file:
                file.write(yaml.dump(obj_as_dict(S.decl)))
            imgui.open_popup("done_save")
        if imgui.begin_popup_modal("done_save")[0]:
            imgui.text(f"Saved file to '{S.save_filename}'")
            if imgui.button(label="OK", width=120, height=0):
                imgui.close_current_popup()
            imgui.set_item_default_focus()
            imgui.end_popup()

        if clicked_quit:
            exit(1)

        menu_size = imgui.get_window_size()
        imgui.end_main_menu_bar()

    # imgui.set_next_window_size(550, 680, condition=imgui.FIRST_USE_EVER)
    io = imgui.get_io()
    imgui.set_next_window_size(io.display_size.x, io.display_size.y - menu_size.y)
    imgui.set_next_window_position(0,menu_size.y)
    flags = imgui.WINDOW_NO_RESIZE | imgui.WINDOW_NO_MOVE | imgui.WINDOW_NO_COLLAPSE | imgui.WINDOW_NO_TITLE_BAR
    # flags |= imgui.WINDOW_NO_BRING_TO_FRONT_ON_FOCUS

    if not imgui.begin(label="SQL Import App", flags=flags):
        # Early out if the window is collapsed, as an optimization.
        imgui.end()
        return
        # sys.exit(1)

    steps = ["0. Introduction",
             "1. Choose Table Types",
             "2. Allocate Columns",
             "3. Finalise Data Types",
             "4. End"]

    imgui.text("Steps: ")
    width = calc_widths(len(steps))
    for step_i,step in enumerate(steps):
        imgui.same_line()

        if step_i != 0:
            imgui.text(">>>>")
            imgui.same_line()

        old_step = S.cur_step
        pop_amount = 0
        if step_i < old_step:
            imgui.push_style_color(imgui.COLOR_BUTTON, *done_step_color)
            imgui.push_style_color(imgui.COLOR_TEXT, *done_step_text_color)
            pop_amount = 2
        if step_i == old_step:
            imgui.push_style_color(imgui.COLOR_BUTTON, *active_step_color)
            pop_amount = 1
        if imgui.button(step):
            S.cur_step = step_i
        imgui.pop_style_color(pop_amount)

    imgui.spacing()
    imgui.separator()
    imgui.separator()
    imgui.spacing()

    if S.cur_step == 0:
        imgui.text_wrapped("""Please walk through the steps of this import one at a time by clicking the buttons at the top of the window. Each step is a "mode" of editing, and presents more detailed choices.

You can always go back to a previous step at any time. Hit save when you are finished.

""")
    elif S.cur_step == 1:
        imgui.text_wrapped("""For each table, select whether each of its rows represents
        a) single entity with attached fields
        b) relation connecting entities from other tables

""")
        col_selecting(S, True)
    elif S.cur_step == 2:
        imgui.text_wrapped("""For each table, choose what the purpose of each column is:
        id:       The unique identifier for this row's entity/relation
        entity:   This column is the unique identifier of an entity from different table
        field:    An attached piece of scalar data (e.g. int, float, string)
        field_on: Advanced only: a piece of scalar data to attach to the connection between two entities.
        ignore:   Don't include this column

""")
        col_selecting(S, False)
    elif S.cur_step == 3:
        imgui.text_wrapped("""You can customize the details from the previous steps, including specific Zef entity/relation names for each column and the data type to use for each column.
        
The checkbox in front of each field should be set to the ID column. This will be removed in future versions and automatically track the "id" column.

""")
        full_col_edit(S)
    elif S.cur_step == 4:
        imgui.text_wrapped(f"""If you are satisfied with the layout, you should now hit "Save" above and then load this into Zef using something like: 
        

from zef import *
from zef.ops import *
from zef.experimental import sql_import

decl = "{S.save_filename}" | load_file | run | get["content"] | collect
g = Graph()
actions = sql_import.import_actions(decl)
actions | transact[g] | run
""")
    else:
        raise Exception("Shouldn't get here")

    imgui.end()


def full_col_edit(S):
    ent_opts = []
    for d in S.decl.definitions:
        if d.kind == "entity":
            ent_opts += [d.ET]

    for d in S.decl.definitions:
        show = table_header(d, True)

        imgui.push_id(d.tag)
        if show:
            # changed, d.kind = combo_autoindex(
            #     label="Kind",
            #     current=d.kind,
            #     items=["entity", "relation"]
            # )
            changed = False
            if imgui.radio_button("Entity", d.kind == "entity"):
                d.kind = "entity"
                changed = True
            same_line()
            if imgui.radio_button("Relation", d.kind == "relation"):
                d.kind = "relation"
                changed = True
            if changed:
                d.ID_col = get(d, "ID_col", None)
                if d.ID_col is None:
                    # Guess
                    for item in d.cols:
                        if item.purpose == "id":
                            d.ID_col = item.name
                            break
                if d.ID_col is None:
                    for item in d.cols:
                        if item.purpose == "entity":
                            d.ID_col = item.name
                            break
                if d.kind == "entity":
                    d.ET = get(d, "ET", get(d, "RT", pascalcase(d.tag)))
                if d.kind == "relation":
                    d.RT = get(d, "RT", get(d, "ET", pascalcase(d.tag)))
                for key in set(keys(d)).difference(kind_keys[d.kind]):
                    delattr(d, key)

            same_line()
            if d.kind == "entity":
                width = calc_widths(1)
                changed,d.ET = raet_input("ET", d.ET, width)

            if d.kind == "relation":
                width = calc_widths(1)
                changed,d.RT = raet_input("RT", d.RT, width)
                
            
            field_on_opts_first = [item.name for item in d.cols if item.purpose == "entity" and item.name != get(d, "ID_col", None)]
            field_on_opts_second = [item.name for item in d.cols if item.name not in field_on_opts_first and item.name != get(d, "ID_col", None)]
            for item in d.cols:
                tree_name = f"{item.purpose.upper()}: {item.name}"
                if item.purpose == "entity":
                    tree_rest = f"ET.{item.ET} using RT.{item.RT}"
                elif item.purpose == "field":
                    tree_rest = f"AET.{item.data_type} using RT.{item.RT}"
                elif item.purpose == "field_on":
                    rt_text = get_col(d.cols, item.target)
                    if rt_text is None:
                        rt_text = "ERROR"
                    else:
                        rt_text = f"RT.{rt_text.RT}"
                    tree_rest = f"AET.{item.data_type} from {item.target} ({rt_text}) using RT.{item.RT}"
                else:
                    tree_rest = ""

                if d.kind == "entity" or d.kind == "relation":
                    checked = d.ID_col == item.name
                    changed,checked = imgui.checkbox(f"###{item.name}Checkbox", checked)
                    if changed:
                        if d.kind == "entity":
                            d.ID_col = item.name
                        elif d.kind == "relation":
                            if checked:
                                d.ID_col = item.name
                            else:
                                d.ID_col = None
                    same_line(5)

                if item.purpose == "entity":
                    color = entity_foreground
                elif item.purpose in ["source", "target", "id"]:
                    color = special_foreground
                else:
                    color = (1.0, 1.0, 1.0)
                imgui.push_style_color(imgui.COLOR_TEXT, *color)
                opened = imgui.tree_node(f"{tree_name} -- {tree_rest}" + "###" + item.name)
                imgui.pop_style_color()
                if opened:
                    with ItemWidth(100):
                        changed, new_purpose = combo_autoindex(
                            label="###Purpose",
                            current=item.purpose,
                            items=purpose_opts
                        )
                    if changed:
                        # We need to update the layout of the struct
                        change_col_to_purpose(item, new_purpose)

                    if item.purpose == "entity":
                        imgui.same_line()
                        width = calc_widths(2)
                        # changed, item.ET = raet_input("ET", item.ET, width)
                        changed, item.ET = entity_selector((d.tag, item.name), item.ET, width, ent_opts)
                        same_line()
                        label = "RT"
                        changed, item.RT = raet_input(label, item.RT, width)
                    elif item.purpose == "field":
                        same_line()
                        width = calc_widths(2)
                        changed, item.RT = raet_input("RT", item.RT, width)
                        same_line()
                        item.data_type = data_type_selector(item.data_type, width)
                    elif item.purpose == "field_on":
                        same_line()
                        width = calc_widths(3)
                        changed, item.RT = raet_input("RT", item.RT, width)
                        same_line()
                        with ItemWidth(width, True):
                            changed, new_target = combo_autoindex(label="Target", current=item.target, items=field_on_opts_first + ["---"] + field_on_opts_second)
                            if changed and new_target != "---":
                                item.target = new_target
                        same_line()
                        item.data_type = data_type_selector(item.data_type, width)
                    elif item.purpose == "ignore":
                        pass
                    elif item.purpose in ["source", "target"]:
                        imgui.same_line()
                        width = calc_widths(1)
                        changed, item.ET = entity_selector((d.tag, item.name), item.ET, width, ent_opts)
                    elif item.purpose == "id":
                        same_line()
                        width = calc_widths(2)
                        changed, item.RT = raet_input("RT", item.RT, width)
                        same_line()
                        item.data_type = data_type_selector(item.data_type, width)
                    else:
                        raise NotImplementedError()
                    imgui.tree_pop()

        imgui.pop_id()



def col_selecting(S, just_tables):
    for d in S.decl.definitions:
        show = table_header(d, False)

        if show:
            imgui.push_id(d.tag)
            changed = False
            if imgui.radio_button("Entity (?)", d.kind == "entity"):
                d.kind = "entity"
                changed = True
            simple_tooltip("Each row in the table will be turned into a Zef entity.")
            same_line()
            if imgui.radio_button("Relation (?)", d.kind == "relation"):
                d.kind = "relation"
                changed = True
            simple_tooltip("Each row in the table represents the relation that connects two entities. These two entities should be described in other tables.")
            if changed:
                d.ID_col = get(d, "ID_col", None)
                if d.ID_col is None:
                    # Guess
                    for item in d.cols:
                        if item.purpose == "id":
                            d.ID_col = item.name
                            break
                if d.ID_col is None:
                    for item in d.cols:
                        if item.purpose == "entity":
                            d.ID_col = item.name
                            break
                if d.kind == "entity":
                    d.ET = get(d, "ET", get(d, "RT", pascalcase(d.tag)))
                if d.kind == "relation":
                    d.RT = get(d, "RT", get(d, "ET", pascalcase(d.tag)))
                for key in set(keys(d)).difference(kind_keys[d.kind]):
                    delattr(d, key)

            all_col_names = [x.name for x in d.cols]
            if d.kind == "entity":
                special_cols = ["id"]
                width = calc_widths(1) - default_label_width
                changed,d.ET = raet_input("ET", d.ET, width, "Zef name")
                simple_tooltip("Each row will generate entities of this ET in the Zef graph.")

            if d.kind == "relation":
                special_cols = ["id", "source", "target"]
                width = calc_widths(1) - default_label_width
                changed,d.RT = raet_input("RT", d.RT, width, "Zef name")
                simple_tooltip("Each row will generate relations of this RT in the Zef graph.")

            if not just_tables:
                imgui.separator()
                imgui.text("Column designations")
                imgui.separator()
                categories = [("field", "Simple data to attach to this entity/relation"),
                            ("entity", "Indicate this column is an ID for a different entity and will create a relation between that entity and this row."),
                            ("field_on", "Only for advanced use. This attaches data to a relation created from a \"entity\" column of this table."),
                            ("ignore", "Columns to ignore")]
                col_width = calc_widths(len(categories) + 1)

                def get_category_length(cat):
                    purpose,title = cat
                    return d.cols | filter[lambda x: x.purpose == purpose] | length | collect
                max_length = categories | map[get_category_length] | max | collect
                height = max(100, (max_length+2) * imgui.get_text_line_height_with_spacing())
                # First the special columns
                imgui.begin_child(f"{d.tag}special", col_width, height, flags=imgui.WINDOW_NO_SCROLL_WITH_MOUSE)
                imgui.text("Special columns")
                imgui.separator()
                for col_s in special_cols:
                    imgui.text(col_s)
                    imgui.same_line()
                    item = get_col_name_from_purpose(d.cols, col_s)
                    item_s = "<none>" if item is None else item
                    width = calc_widths(1)
                    imgui.selectable(item_s + f"###{col_s}selectable", True, width=width)
                    if item is not None:
                        if imgui.begin_drag_drop_source():
                            imgui.set_drag_drop_payload(f"{d.tag} drag", item.encode("ascii"))
                            imgui.text(item)
                            imgui.end_drag_drop_source()
                    if imgui.begin_drag_drop_target():
                        payload = imgui.accept_drag_drop_payload(f"{d.tag} drag")
                        if payload is not None:
                            new_item = payload.decode("ascii")
                            col = get_col(d.cols, new_item)
                            change_col_to_purpose(col, col_s)
                            # Reset the old one's purpose to be a generic field
                            if item is not None:
                                col = get_col(d.cols, item)
                                change_col_to_purpose(col, "field")
                        imgui.end_drag_drop_target()

                    if col_s == "id":
                        tool_s = "This sets the column to be how other tables can link to this row. An entity must have an ID but it is optional for a relation."
                    elif col_s == "source":
                        tool_s = "The column from which the source entity will be drawn. This column should match another table's ID column. This is like a SQL foreign key."
                    elif col_s == "target":
                        tool_s = "The column from which the target entity will be drawn. This column should match another table's ID column. This is like a SQL foreign key."
                    simple_tooltip(tool_s)

                    imgui.spacing()
                    imgui.spacing()
                    imgui.spacing()
                imgui.end_child()
                imgui.same_line()

                for cat_i,(cat,title) in enumerate(categories):
                    if cat_i > 0:
                        imgui.same_line()
                    imgui.begin_child(f"{d.tag}{cat}", col_width, height, border=True, flags=imgui.WINDOW_NO_SCROLL_WITH_MOUSE)
                    imgui.text(f'"{cat}"')
                    simple_tooltip(title)
                    imgui.separator()
                    items = [x.name for x in d.cols if x.purpose == cat]
                    for item in items:
                        imgui.selectable(item, False)
                        if imgui.begin_drag_drop_source():
                            imgui.set_drag_drop_payload(f"{d.tag} drag", item.encode("ascii"))
                            imgui.text(item)
                            imgui.end_drag_drop_source()
                    imgui.end_child()
                    if imgui.begin_drag_drop_target():
                        payload = imgui.accept_drag_drop_payload(f"{d.tag} drag")
                        if payload is not None:
                            item = payload.decode("ascii")
                            col = get_col(d.cols, item)
                            change_col_to_purpose(col, cat)
                        imgui.end_drag_drop_target()

            imgui.dummy(0, 20)
            imgui.pop_id()

def table_header(d, show_details):
    if d.kind == "entity":
        color = entity_background
        name_def = f"ENTITY: ET.{d.ET}"
    elif d.kind == "relation":
        color = relation_background
        if show_details:
            source_cols = [x for x in d.cols if x.purpose == "source"]
            if len(source_cols) != 1:
                source_name = "UNKNOWN"
            else:
                source_name = "ET." + source_cols[0].ET
            target_cols = [x for x in d.cols if x.purpose == "target"]
            if len(target_cols) != 1:
                target_name = "UNKNOWN"
            else:
                target_name = "ET." + target_cols[0].ET
            name_def = f"RELATION: {source_name}--RT.{d.RT}->{target_name}"
        else:
            name_def = f"RELATION: RT.{d.RT}"
    else:
        color = (1.0, 1.0, 1.0)
        name_def = "UNKNOWN"
    name = f"Table {d.tag}, {d.data_source.filename}: {name_def}"
    # imgui.push_style_color(imgui.COLOR_TEXT, *color)
    imgui.push_style_color(imgui.COLOR_HEADER, *color)
    show,_ = imgui.collapsing_header(name + "###" + d.tag, flags=imgui.TREE_NODE_DEFAULT_OPEN)
    imgui.pop_style_color()
    return show


##############################
# * Validation
#----------------------------

def todo():
    # TODO: entities can't have ID purposes
    # TODO: Exactly one source/target in a relation kind
    # TODO: Maximum one ID purpose in relation kind
    # TODO: Exactly one ID purpose 
    pass



################################
# * ImGUI helpers
#------------------------------



default_spacing = 20
default_label_width = 100

def calc_widths(N, spacing=default_spacing, full_width=None):
    if full_width is None:
        full_width = imgui.get_content_region_available_width()
    return (full_width - (N-1)*default_spacing) / N

def same_line(spacing=default_spacing):
    imgui.same_line(spacing=spacing)

from contextlib import contextmanager
@contextmanager
def ItemWidth(w, label_w=0):
    if label_w is True:
        label_w = default_label_width
    imgui.push_item_width(w-label_w)
    yield
    imgui.pop_item_width()
    imgui.set_cursor_pos_x(imgui.get_cursor_pos_x() + label_w)
    


def combo_autoindex(label, current, items, **kwds):
    if current in items:
        index = items.index(current)
    else:
        index = -1

    changed, new_val = imgui.combo(
        label=label,
        current=index,
        items=items,
        **kwds
    )
    return changed, items[new_val]

def data_type_selector(val, full_width):
    imgui.push_id("DataTypeSelector")
    if '.' in val:
        parts = val.split('.')
        typ = parts[0]
        subtyp = '.'.join(parts[1:])
    else:
        typ = val
        subtyp = None

    if subtyp is None:
        N = 1
    else:
        N = 2

    start_x = imgui.get_cursor_pos_x()
    spacing = 5

    with ItemWidth(0):
        imgui.text("AET.")
    same_line(0)

    with ItemWidth(150):
        changed, new_opt = combo_autoindex(
            label="###Combo",
            current=typ,
            items=["Bool", "Int", "Float", "String", "Time", "QuantityFloat", "QuantityInt", "Enum"]
        )
    if subtyp is not None:
        same_line(0)
        with ItemWidth(0):
            imgui.text(".")
        same_line(0)
        leftover = full_width - (imgui.get_cursor_pos_x() - start_x)
        with ItemWidth(leftover):
            changed, new_subtyp = imgui.input_text(label="###SubType", value=subtyp, buffer_length=50)
    else:
        new_subtyp = None

    if new_opt in ["QuantityFloat", "QuantityInt", "Enum"]:
        if new_subtyp is None:
            new_subtyp = ""
    else:
        new_subtyp = None

    imgui.pop_id()

    if new_subtyp is None:
        return new_opt
    else:
        return f"{new_opt}.{new_subtyp}"


def raet_input(token, value, width, label=""):
    imgui.begin_group()
    imgui.text(token + ".")
    used_width = imgui.get_item_rect_size()[0]
    same_line(0)
    with ItemWidth(width - used_width):
        changed, value = imgui.input_text(label=f"{label}###{token}{label}", value=value, buffer_length=50)
    imgui.end_group()
    return changed, value


last_was_popup_active = {}
def entity_selector(id, value, width, ent_opts):
    # Provides a dropdown selectable for choosing an entity from current ones in the list
    something_clicked = False
    
    # changed,value = imgui.input_text(label="TESTING", value=value, buffer_length=50)
    change_bg = value not in ent_opts
    if change_bg:
        imgui.push_style_color(imgui.COLOR_FRAME_BACKGROUND, 0.5,0.1,0.1)
    changed,value = raet_input("ET", value, width)
    input_active = imgui.is_item_active() or changed
    input_focused = imgui.is_item_focused()
    if change_bg:
        imgui.pop_style_color()
    if input_active or last_was_popup_active.setdefault(id, False):
        imgui.set_next_window_position(imgui.get_item_rect_min().x, imgui.get_item_rect_max().y)
        flags = imgui.WINDOW_NO_TITLE_BAR | imgui.WINDOW_NO_RESIZE | imgui.WINDOW_NO_MOVE | imgui.WINDOW_ALWAYS_AUTO_RESIZE | imgui.WINDOW_ALWAYS_VERTICAL_SCROLLBAR
        imgui.begin("popup###" + str(id), flags=flags)
        last_was_popup_active[id] = imgui.is_window_focused()
        if input_active:
            imgui.set_window_focus()
        for x in ent_opts:
            clicked, _ = imgui.selectable(x, False)
            if clicked:
                changed = True
                value = x
                something_clicked = True
        # imgui.end_tooltip()
        imgui.end()
    else:
        last_was_popup_active[id] = False

    if something_clicked:
        last_was_popup_active[id] = False

    return changed,value

def simple_tooltip(x):
    if imgui.is_item_hovered():
        imgui.begin_tooltip()
        imgui.push_text_wrap_position(400)
        imgui.text_wrapped(x)
        imgui.end_tooltip()
        # imgui.set_tooltip(x)